import os

import dbt.exceptions
import dbt.flags
import dbt.include
import dbt.utils
import dbt.hooks
import dbt.clients.jinja
import dbt.context.parser

from dbt.include.global_project import PROJECT_NAME as GLOBAL_PROJECT_NAME
from dbt.utils import coalesce
from dbt.logger import GLOBAL_LOGGER as logger
from dbt.contracts.graph.parsed import ParsedNode
from dbt.parser.source_config import SourceConfig
from dbt import deprecations


class BaseParser(object):
    def __init__(self, root_project_config, all_projects):
        self.root_project_config = root_project_config
        self.all_projects = all_projects

    @property
    def default_schema(self):
        return getattr(self.root_project_config.credentials, 'schema',
                       'public')

    @property
    def default_database(self):
        return getattr(self.root_project_config.credentials, 'database', 'dbt')

    def load_and_parse(self, *args, **kwargs):
        raise dbt.exceptions.NotImplementedException("Not implemented")

    @classmethod
    def get_path(cls, resource_type, package_name, resource_name):
        """Returns a unique identifier for a resource"""

        return "{}.{}.{}".format(resource_type, package_name, resource_name)

    @classmethod
    def get_fqn(cls, node, package_project_config, extra=[]):
        parts = dbt.utils.split_path(node.path)
        name, _ = os.path.splitext(parts[-1])
        fqn = ([package_project_config.project_name] +
               parts[:-1] +
               extra +
               [name])

        return fqn


class MacrosKnownParser(BaseParser):
    def __init__(self, root_project_config, all_projects, macro_manifest):
        super(MacrosKnownParser, self).__init__(
            root_project_config=root_project_config,
            all_projects=all_projects
        )
        self.macro_manifest = macro_manifest
        self._get_schema_func = None
        self._get_alias_func = None

    def get_schema_func(self):
        """The get_schema function is set by a few different things:
            - if there is a 'generate_schema_name' macro in the root project,
                it will be used.
            - if that does not exist but there is a 'generate_schema_name'
                macro in the 'dbt' internal project, that will be used
            - if neither of those exist (unit tests?), a function that returns
                the 'default schema' as set in the root project's 'credentials'
                is used
        """
        if self._get_schema_func is not None:
            return self._get_schema_func

        get_schema_macro = self.macro_manifest.find_macro_by_name(
            'generate_schema_name',
            self.root_project_config.project_name
        )
        if get_schema_macro is None:
            get_schema_macro = self.macro_manifest.find_macro_by_name(
                'generate_schema_name',
                GLOBAL_PROJECT_NAME
            )
        # this is only true in tests!
        if get_schema_macro is None:
            def get_schema(custom_schema_name=None, node=None):
                return self.default_schema
        else:
            root_context = dbt.context.parser.generate_macro(
                get_schema_macro, self.root_project_config,
                self.macro_manifest
            )
            get_schema = get_schema_macro.generator(root_context)

        self._get_schema_func = get_schema
        return self._get_schema_func

    def get_alias_func(self):
        """The get_alias function is set by a few different things:
            - if there is a 'generate_alias_name' macro in the root project,
                it will be used.
            - if that does not exist but there is a 'generate_alias_name'
                macro in the 'dbt' internal project, that will be used
            - if neither of those exist (unit tests?), a function that returns
                the 'default alias' as set in the model's filename or alias
                configuration.
        """
        if self._get_alias_func is not None:
            return self._get_alias_func

        get_alias_macro = self.macro_manifest.find_macro_by_name(
            'generate_alias_name',
            self.root_project_config.project_name
        )
        if get_alias_macro is None:
            get_alias_macro = self.macro_manifest.find_macro_by_name(
                'generate_alias_name',
                GLOBAL_PROJECT_NAME
            )

        # the generate_alias_name macro might not exist
        if get_alias_macro is None:
            def get_alias(custom_alias_name, node):
                if custom_alias_name is None:
                    return node.name
                else:
                    return custom_alias_name
        else:
            root_context = dbt.context.parser.generate_macro(
                get_alias_macro, self.root_project_config,
                self.macro_manifest
            )
            get_alias = get_alias_macro.generator(root_context)

        self._get_alias_func = get_alias
        return self._get_alias_func

    def _build_intermediate_node_dict(self, config, node_dict, node_path,
                                      package_project_config, tags, fqn,
                                      agate_table, snapshot_config,
                                      column_name):
        """Update the unparsed node dictionary and build the basis for an
        intermediate ParsedNode that will be passed into the renderer
        """
        # because this takes and returns dicts, subclasses can safely override
        # this and mutate its results using super() both before and after.
        if agate_table is not None:
            node_dict['agate_table'] = agate_table

        # Set this temporarily. Not the full config yet (as config() hasn't
        # been called from jinja yet). But the Var() call below needs info
        # about project level configs b/c they might contain refs.
        # TODO: Restructure this?
        config_dict = coalesce(snapshot_config, {})
        config_dict.update(config.config)

        empty = (
            'raw_sql' in node_dict and len(node_dict['raw_sql'].strip()) == 0
        )

        node_dict.update({
            'refs': [],
            'sources': [],
            'depends_on': {
                'nodes': [],
                'macros': [],
            },
            'unique_id': node_path,
            'empty': empty,
            'fqn': fqn,
            'tags': tags,
            'config': config_dict,
            # Set these temporarily so get_rendered() has access to a schema,
            # database, and alias.
            'schema': self.default_schema,
            'database': self.default_database,
            'alias': node_dict.get('name'),
        })

        # if there's a column, it should end up part of the ParsedNode
        if column_name is not None:
            node_dict['column_name'] = column_name

        return node_dict

    def _render_with_context(self, parsed_node, config):
        """Given the parsed node and a SourceConfig to use during parsing,
        render the node's sql wtih macro capture enabled.

        Note: this mutates the config object when config() calls are rendered.
        """
        context = dbt.context.parser.generate(
            parsed_node,
            self.root_project_config,
            self.macro_manifest,
            config)

        dbt.clients.jinja.get_rendered(
            parsed_node.raw_sql, context, parsed_node.to_shallow_dict(),
            capture_macros=True)

    def _update_parsed_node_info(self, parsed_node, config):
        """Given the SourceConfig used for parsing and the parsed node,
        generate and set the true values to use, overriding the temporary parse
        values set in _build_intermediate_parsed_node.
        """
        # Special macro defined in the global project. Use the root project's
        # definition, not the current package
        schema_override = config.config.get('schema')
        get_schema = self.get_schema_func()
        try:
            schema = get_schema(schema_override, parsed_node)
        except dbt.exceptions.CompilationException as exc:
            too_many_args = (
                "macro 'dbt_macro__generate_schema_name' takes not more than "
                "1 argument(s)"
            )
            if too_many_args not in str(exc):
                raise
            deprecations.warn('generate-schema-name-single-arg')
            schema = get_schema(schema_override)
        parsed_node.schema = schema.strip()

        alias_override = config.config.get('alias')
        get_alias = self.get_alias_func()
        parsed_node.alias = get_alias(alias_override, parsed_node).strip()

        parsed_node.database = config.config.get(
            'database', self.default_database
        ).strip()

        # Set tags on node provided in config blocks
        model_tags = config.config.get('tags', [])
        parsed_node.tags.extend(model_tags)

        # Overwrite node config
        config_dict = parsed_node.get('config', {})
        config_dict.update(config.config)
        parsed_node.config = config_dict

        for hook_type in dbt.hooks.ModelHookType.Both:
            parsed_node.config[hook_type] = dbt.hooks.get_hooks(parsed_node,
                                                                hook_type)

    def parse_node(self, node, node_path, package_project_config, tags=None,
                   fqn_extra=None, fqn=None, agate_table=None,
                   snapshot_config=None, column_name=None):
        """Parse a node, given an UnparsedNode and any other required information.

        agate_table should be set if the node came from a seed file.
        snapshot_config should be set if the node is an Snapshot node.
        column_name should be set if the node is a Test node associated with a
        particular column.
        """
        logger.debug("Parsing {}".format(node_path))

        tags = coalesce(tags, [])
        fqn_extra = coalesce(fqn_extra, [])

        if fqn is None:
            fqn = self.get_fqn(node, package_project_config, fqn_extra)

        config = SourceConfig(
            self.root_project_config,
            package_project_config,
            fqn,
            node.resource_type)

        parsed_dict = self._build_intermediate_node_dict(
            config, node.serialize(), node_path, config, tags, fqn,
            agate_table, snapshot_config, column_name
        )
        parsed_node = ParsedNode(**parsed_dict)

        self._render_with_context(parsed_node, config)
        self._update_parsed_node_info(parsed_node, config)

        parsed_node.validate()

        return parsed_node

    def check_block_parsing(self, name, path, contents):
        """Check if we were able to extract toplevel blocks from the given
        contents. Return True if extraction was successful (no exceptions),
        False if it fails.
        """
        if not dbt.flags.TEST_NEW_PARSER:
            return True
        try:
            dbt.clients.jinja.extract_toplevel_blocks(contents)
        except Exception:
            return False
        return True



from dbt.config.project import Project


class AdapterPlugin(object):
    """Defines the basic requirements for a dbt adapter plugin.

    :param type adapter: An adapter class, derived from BaseAdapter
    :param type credentials: A credentials object, derived from Credentials
    :param str project_name: The name of this adapter plugin's associated dbt
        project.
    :param str include_path: The path to this adapter plugin's root
    :param Optional[List[str]] dependencies: A list of adapter names that this
        adapter depends upon.
    """
    def __init__(self, adapter, credentials, include_path, dependencies=None):
        self.adapter = adapter
        self.credentials = credentials
        self.include_path = include_path
        project = Project.from_project_root(include_path, {})
        self.project_name = project.project_name
        if dependencies is None:
            dependencies = []
        self.dependencies = dependencies



import base64
import os
import re
import time
from abc import abstractmethod
from multiprocessing.dummy import Pool as ThreadPool

from dbt import rpc
from dbt.task.base import ConfiguredTask
from dbt.adapters.factory import get_adapter
from dbt.logger import GLOBAL_LOGGER as logger
from dbt.compat import to_unicode
from dbt.compilation import compile_manifest
from dbt.contracts.graph.manifest import CompileResultNode
from dbt.contracts.results import ExecutionResult
from dbt.loader import GraphLoader

import dbt.exceptions
import dbt.ui.printer
import dbt.utils

import dbt.graph.selector

RESULT_FILE_NAME = 'run_results.json'
MANIFEST_FILE_NAME = 'manifest.json'


def load_manifest(config):
    # performance trick: if the adapter has a manifest loaded, use that to
    # avoid parsing internal macros twice.
    internal_manifest = get_adapter(config).check_internal_manifest()
    manifest = GraphLoader.load_all(config,
                                    internal_manifest=internal_manifest)

    manifest.write(os.path.join(config.target_path, MANIFEST_FILE_NAME))
    return manifest


class ManifestTask(ConfiguredTask):
    def __init__(self, args, config):
        super(ManifestTask, self).__init__(args, config)
        self.manifest = None
        self.linker = None

    def _runtime_initialize(self):
        self.manifest = load_manifest(self.config)
        self.linker = compile_manifest(self.config, self.manifest)


class GraphRunnableTask(ManifestTask):
    def __init__(self, args, config):
        super(GraphRunnableTask, self).__init__(args, config)
        self.job_queue = None
        self._flattened_nodes = None

        self.run_count = 0
        self.num_nodes = None
        self.node_results = []
        self._skipped_children = {}
        self._raise_next_tick = None

    def select_nodes(self):
        selector = dbt.graph.selector.NodeSelector(self.linker, self.manifest)
        selected_nodes = selector.select(self.build_query())
        return selected_nodes

    def _runtime_initialize(self):
        super(GraphRunnableTask, self)._runtime_initialize()
        selected_nodes = self.select_nodes()
        self.job_queue = self.linker.as_graph_queue(self.manifest,
                                                    selected_nodes)

        # we use this a couple times. order does not matter.
        self._flattened_nodes = [
            self.manifest.nodes[uid] for uid in selected_nodes
        ]

        self.num_nodes = len([
            n for n in self._flattened_nodes
            if not n.is_ephemeral_model
        ])

    def raise_on_first_error(self):
        return False

    def build_query(self):
        raise dbt.exceptions.NotImplementedException('Not Implemented')

    def get_runner_type(self):
        raise dbt.exceptions.NotImplementedException('Not Implemented')

    def result_path(self):
        return os.path.join(self.config.target_path, RESULT_FILE_NAME)

    def get_runner(self, node):
        adapter = get_adapter(self.config)

        if node.is_ephemeral_model:
            run_count = 0
            num_nodes = 0
        else:
            self.run_count += 1
            run_count = self.run_count
            num_nodes = self.num_nodes

        cls = self.get_runner_type()
        return cls(self.config, adapter, node, run_count, num_nodes)

    def call_runner(self, runner):
        # TODO: create+enforce an actual contracts for what `result` is instead
        # of the current free-for-all
        result = runner.run_with_hooks(self.manifest)
        if result.error is not None and self.raise_on_first_error():
            # if we raise inside a thread, it'll just get silently swallowed.
            # stash the error message we want here, and it will check the
            # next 'tick' - should be soon since our thread is about to finish!
            self._raise_next_tick = result.error

        return result

    def _submit(self, pool, args, callback):
        """If the caller has passed the magic 'single-threaded' flag, call the
        function directly instead of pool.apply_async. The single-threaded flag
         is intended for gathering more useful performance information about
        what appens beneath `call_runner`, since python's default profiling
        tools ignore child threads.

        This does still go through the callback path for result collection.
        """
        if self.config.args.single_threaded:
            callback(self.call_runner(*args))
        else:
            pool.apply_async(self.call_runner, args=args, callback=callback)

    def _raise_set_error(self):
        if self._raise_next_tick is not None:
            raise dbt.exceptions.RuntimeException(self._raise_next_tick)

    def run_queue(self, pool):
        """Given a pool, submit jobs from the queue to the pool.
        """
        def callback(result):
            """Note: mark_done, at a minimum, must happen here or dbt will
            deadlock during ephemeral result error handling!
            """
            self._handle_result(result)
            self.job_queue.mark_done(result.node.unique_id)

        while not self.job_queue.empty():
            node = self.job_queue.get()
            self._raise_set_error()
            runner = self.get_runner(node)
            # we finally know what we're running! Make sure we haven't decided
            # to skip it due to upstream failures
            if runner.node.unique_id in self._skipped_children:
                cause = self._skipped_children.pop(runner.node.unique_id)
                runner.do_skip(cause=cause)
            args = (runner,)
            self._submit(pool, args, callback)

        # block on completion
        self.job_queue.join()
        # if an error got set during join(), raise it.
        self._raise_set_error()

        return

    def _handle_result(self, result):
        """Mark the result as completed, insert the `CompiledResultNode` into
        the manifest, and mark any descendants (potentially with a 'cause' if
        the result was an ephemeral model) as skipped.
        """
        is_ephemeral = result.node.is_ephemeral_model
        if not is_ephemeral:
            self.node_results.append(result)

        node = CompileResultNode(**result.node)
        node_id = node.unique_id
        self.manifest.nodes[node_id] = node

        if result.error is not None:
            if is_ephemeral:
                cause = result
            else:
                cause = None
            self._mark_dependent_errors(node_id, result, cause)

    def execute_nodes(self):
        num_threads = self.config.threads
        target_name = self.config.target_name

        text = "Concurrency: {} threads (target='{}')"
        concurrency_line = text.format(num_threads, target_name)
        dbt.ui.printer.print_timestamped_line(concurrency_line)
        dbt.ui.printer.print_timestamped_line("")

        pool = ThreadPool(num_threads)
        try:
            self.run_queue(pool)

        except KeyboardInterrupt:
            pool.close()
            pool.terminate()

            adapter = get_adapter(self.config)

            if not adapter.is_cancelable():
                msg = ("The {} adapter does not support query "
                       "cancellation. Some queries may still be "
                       "running!".format(adapter.type()))

                yellow = dbt.ui.printer.COLOR_FG_YELLOW
                dbt.ui.printer.print_timestamped_line(msg, yellow)
                raise

            for conn_name in adapter.cancel_open_connections():
                dbt.ui.printer.print_cancel_line(conn_name)

            pool.join()

            dbt.ui.printer.print_run_end_messages(self.node_results,
                                                  early_exit=True)

            raise

        pool.close()
        pool.join()

        return self.node_results

    def _mark_dependent_errors(self, node_id, result, cause):
        for dep_node_id in self.linker.get_dependent_nodes(node_id):
            self._skipped_children[dep_node_id] = cause

    def before_hooks(self, adapter):
        pass

    def before_run(self, adapter, selected_uids):
        pass

    def after_run(self, adapter, results):
        pass

    def after_hooks(self, adapter, results, elapsed):
        pass

    def execute_with_hooks(self, selected_uids):
        adapter = get_adapter(self.config)
        try:
            self.before_hooks(adapter)
            started = time.time()
            self.before_run(adapter, selected_uids)
            res = self.execute_nodes()
            self.after_run(adapter, res)
            elapsed = time.time() - started
            self.after_hooks(adapter, res, elapsed)

        finally:
            adapter.cleanup_connections()

        result = self.get_result(
            results=res,
            elapsed_time=elapsed,
            generated_at=dbt.utils.timestring()
        )
        return result

    def run(self):
        """
        Run dbt for the query, based on the graph.
        """
        self._runtime_initialize()

        if len(self._flattened_nodes) == 0:
            logger.warning("WARNING: Nothing to do. Try checking your model "
                           "configs and model specification args")
            return []
        else:
            logger.info("")

        selected_uids = frozenset(n.unique_id for n in self._flattened_nodes)
        result = self.execute_with_hooks(selected_uids)

        result.write(self.result_path())

        self.task_end_messages(result.results)
        return result.results

    def interpret_results(self, results):
        if results is None:
            return False

        failures = [r for r in results if r.error or r.failed]
        return len(failures) == 0

    def get_model_schemas(self, selected_uids):
        schemas = set()
        for node in self.manifest.nodes.values():
            if node.unique_id not in selected_uids:
                continue
            if node.is_refable and not node.is_ephemeral:
                schemas.add((node.database, node.schema))

        return schemas

    def create_schemas(self, adapter, selected_uids):
        required_schemas = self.get_model_schemas(selected_uids)

        # Snowflake needs to issue a "use {schema}" query, where schema
        # is the one defined in the profile. Create this schema if it
        # does not exist, otherwise subsequent queries will fail. Generally,
        # dbt expects that this schema will exist anyway.
        required_schemas.add(
            (self.config.credentials.database, self.config.credentials.schema)
        )

        required_databases = set(db for db, _ in required_schemas)

        existing_schemas = set()
        for db in required_databases:
            existing_schemas.update((db, s) for s in adapter.list_schemas(db))

        for database, schema in (required_schemas - existing_schemas):
            adapter.create_schema(database, schema)

    def get_result(self, results, elapsed_time, generated_at):
        return ExecutionResult(
            results=results,
            elapsed_time=elapsed_time,
            generated_at=generated_at
        )

    def task_end_messages(self, results):
        dbt.ui.printer.print_run_end_messages(results)


class RemoteCallable(object):
    METHOD_NAME = None
    is_async = False

    @abstractmethod
    def handle_request(self, **kwargs):
        raise dbt.exceptions.NotImplementedException(
            'from_kwargs not implemented'
        )

    def decode_sql(self, sql):
        """Base64 decode a string. This should only be used for sql in calls.

        :param str sql: The base64 encoded form of the original utf-8 string
        :return str: The decoded utf-8 string
        """
        # JSON is defined as using "unicode", we'll go a step further and
        # mandate utf-8 (though for the base64 part, it doesn't really matter!)
        base64_sql_bytes = to_unicode(sql).encode('utf-8')
        # in python3.x you can pass `validate=True` to b64decode to get this
        # behavior.
        if not re.match(b'^[A-Za-z0-9+/]*={0,2}$', base64_sql_bytes):
            self.raise_invalid_base64(sql)

        try:
            sql_bytes = base64.b64decode(base64_sql_bytes)
        except ValueError:
            self.raise_invalid_base64(sql)

        return sql_bytes.decode('utf-8')

    @staticmethod
    def raise_invalid_base64(sql):
        raise rpc.invalid_params(
            data={
                'message': 'invalid base64-encoded sql input',
                'sql': str(sql),
            }
        )



import os

import dbt.flags
import dbt.clients.agate_helper
import dbt.clients.system
import dbt.context.parser
import dbt.contracts.project
import dbt.exceptions

from dbt.node_types import NodeType
from dbt.logger import GLOBAL_LOGGER as logger
from dbt.contracts.graph.unparsed import UnparsedNode
from dbt.parser.base import MacrosKnownParser


class SeedParser(MacrosKnownParser):
    @classmethod
    def parse_seed_file(cls, file_match, root_dir, package_name, should_parse):
        """Parse the given seed file, returning an UnparsedNode and the agate
        table.
        """
        abspath = file_match['absolute_path']
        logger.debug("Parsing {}".format(abspath))
        table_name = os.path.basename(abspath)[:-4]
        node = UnparsedNode(
            path=file_match['relative_path'],
            name=table_name,
            root_path=root_dir,
            resource_type=NodeType.Seed,
            # Give this raw_sql so it conforms to the node spec,
            # use dummy text so it doesn't look like an empty node
            raw_sql='-- csv --',
            package_name=package_name,
            original_file_path=os.path.join(file_match.get('searched_path'),
                                            file_match.get('relative_path')),
        )
        if should_parse:
            try:
                table = dbt.clients.agate_helper.from_csv(abspath)
            except ValueError as e:
                dbt.exceptions.raise_compiler_error(str(e), node)
        else:
            table = dbt.clients.agate_helper.empty_table()
        table.original_abspath = abspath
        return node, table

    def load_and_parse(self, package_name, root_dir, relative_dirs, tags=None):
        """Load and parse seed files in a list of directories. Returns a dict
           that maps unique ids onto ParsedNodes"""

        extension = "[!.#~]*.csv"
        if dbt.flags.STRICT_MODE:
            dbt.contracts.project.ProjectList(**self.all_projects)

        file_matches = dbt.clients.system.find_matching(
            root_dir,
            relative_dirs,
            extension)

        # we only want to parse seeds if we're inside 'dbt seed'
        should_parse = self.root_project_config.args.which == 'seed'

        result = {}
        for file_match in file_matches:
            node, agate_table = self.parse_seed_file(file_match, root_dir,
                                                     package_name,
                                                     should_parse)
            node_path = self.get_path(NodeType.Seed, package_name, node.name)
            parsed = self.parse_node(node, node_path,
                                     self.all_projects.get(package_name),
                                     tags=tags, agate_table=agate_table)
            result[node_path] = parsed

        return result



from __future__ import print_function

import json

from dbt.task.runnable import GraphRunnableTask, ManifestTask
from dbt.node_types import NodeType
import dbt.exceptions
from dbt.logger import GLOBAL_LOGGER as logger
from dbt.logger import log_to_stderr


class ListTask(GraphRunnableTask):
    DEFAULT_RESOURCE_VALUES = frozenset((
        NodeType.Model,
        NodeType.Snapshot,
        NodeType.Seed,
        NodeType.Test,
        NodeType.Source,
    ))
    ALL_RESOURCE_VALUES = DEFAULT_RESOURCE_VALUES | frozenset((
        NodeType.Analysis,
    ))
    ALLOWED_KEYS = frozenset((
        'alias',
        'name',
        'package_name',
        'depends_on',
        'tags',
        'config',
        'resource_type',
        'source_name',
    ))

    def __init__(self, args, config):
        super(ListTask, self).__init__(args, config)
        self.args.single_threaded = True
        if self.args.models:
            if self.args.select:
                raise dbt.exceptions.RuntimeException(
                    '"models" and "select" are mutually exclusive arguments'
                )
            if self.args.resource_types:
                raise dbt.exceptions.RuntimeException(
                    '"models" and "resource_type" are mutually exclusive '
                    'arguments'
                )

    @classmethod
    def pre_init_hook(cls):
        """A hook called before the task is initialized."""
        log_to_stderr(logger)

    def _iterate_selected_nodes(self):
        nodes = sorted(self.select_nodes())
        if not nodes:
            logger.warning('No nodes selected!')
            return
        for node in nodes:
            yield self.manifest.nodes[node]

    def generate_selectors(self):
        for node in self._iterate_selected_nodes():
            selector = '.'.join(node.fqn)
            if node.resource_type == NodeType.Source:
                yield 'source:{}'.format(selector)
            else:
                yield selector

    def generate_names(self):
        for node in self._iterate_selected_nodes():
            if node.resource_type == NodeType.Source:
                yield '{0.source_name}.{0.name}'.format(node)
            else:
                yield node.name

    def generate_json(self):
        for node in self._iterate_selected_nodes():
            yield json.dumps({
                k: v
                for k, v in node.serialize().items()
                if k in self.ALLOWED_KEYS
            })

    def generate_paths(self):
        for node in self._iterate_selected_nodes():
            yield node.get('original_file_path')

    def run(self):
        ManifestTask._runtime_initialize(self)
        output = self.config.args.output
        if output == 'selector':
            generator = self.generate_selectors
        elif output == 'name':
            generator = self.generate_names
        elif output == 'json':
            generator = self.generate_json
        elif output == 'path':
            generator = self.generate_paths
        else:
            raise dbt.exceptions.IternalException(
                'Invalid output {}'.format(output)
            )
        for result in generator():
            self.node_results.append(result)
            print(result)
        return self.node_results

    @property
    def resource_types(self):
        if self.args.models:
            return [NodeType.Model]

        values = set(self.config.args.resource_types)
        if not values:
            return list(self.DEFAULT_RESOURCE_VALUES)

        if 'default' in values:
            values.remove('default')
            values.update(self.DEFAULT_RESOURCE_VALUES)
        if 'all' in values:
            values.remove('all')
            values.update(self.ALL_RESOURCE_VALUES)
        return list(values)

    @property
    def selector(self):
        if self.args.models:
            return self.args.models
        else:
            return self.args.select

    def build_query(self):
        return {
            "include": self.selector,
            "exclude": self.args.exclude,
            "resource_types": self.resource_types,
            "tags": [],
        }

    def interpret_results(self, results):
        return bool(results)



import os
import signal
import threading

from dbt.adapters.factory import get_adapter
from dbt.clients.jinja import extract_toplevel_blocks
from dbt.compilation import compile_manifest
from dbt.loader import load_all_projects
from dbt.node_runners import CompileRunner, RPCCompileRunner
from dbt.node_types import NodeType
from dbt.parser.analysis import RPCCallParser
from dbt.parser.macros import MacroParser
from dbt.parser.util import ParserUtils
import dbt.ui.printer
from dbt.logger import RPC_LOGGER as rpc_logger

from dbt.task.runnable import GraphRunnableTask, RemoteCallable


class CompileTask(GraphRunnableTask):

    def raise_on_first_error(self):
        return True

    def build_query(self):
        return {
            "include": self.args.models,
            "exclude": self.args.exclude,
            "resource_types": NodeType.executable(),
            "tags": [],
        }

    def get_runner_type(self):
        return CompileRunner

    def task_end_messages(self, results):
        dbt.ui.printer.print_timestamped_line('Done.')


class RemoteCompileTask(CompileTask, RemoteCallable):
    METHOD_NAME = 'compile'

    def __init__(self, args, config, manifest):
        super(RemoteCompileTask, self).__init__(args, config)
        self._base_manifest = manifest.deepcopy(config=config)

    def get_runner_type(self):
        return RPCCompileRunner

    def runtime_cleanup(self, selected_uids):
        """Do some pre-run cleanup that is usually performed in Task __init__.
        """
        self.run_count = 0
        self.num_nodes = len(selected_uids)
        self.node_results = []
        self._skipped_children = {}
        self._skipped_children = {}
        self._raise_next_tick = None

    def _extract_request_data(self, data):
        data = self.decode_sql(data)
        macro_blocks = []
        data_chunks = []
        for block in extract_toplevel_blocks(data):
            if block.block_type_name == 'macro':
                macro_blocks.append(block.full_block)
            else:
                data_chunks.append(block.full_block)
        macros = '\n'.join(macro_blocks)
        sql = ''.join(data_chunks)
        return sql, macros

    def _get_exec_node(self, name, sql, macros):
        request_path = os.path.join(self.config.target_path, 'rpc', name)
        all_projects = load_all_projects(self.config)
        macro_overrides = {}
        sql, macros = self._extract_request_data(sql)

        if macros:
            macro_parser = MacroParser(self.config, all_projects)
            macro_overrides.update(macro_parser.parse_macro_file(
                macro_file_path='from remote system',
                macro_file_contents=macros,
                root_path=request_path,
                package_name=self.config.project_name,
                resource_type=NodeType.Macro
            ))

        self._base_manifest.macros.update(macro_overrides)
        rpc_parser = RPCCallParser(
            self.config,
            all_projects=all_projects,
            macro_manifest=self._base_manifest
        )

        node_dict = {
            'name': name,
            'root_path': request_path,
            'resource_type': NodeType.RPCCall,
            'path': name + '.sql',
            'original_file_path': 'from remote system',
            'package_name': self.config.project_name,
            'raw_sql': sql,
        }

        unique_id, node = rpc_parser.parse_sql_node(node_dict)
        self.manifest = ParserUtils.add_new_refs(
            manifest=self._base_manifest,
            current_project=self.config,
            node=node,
            macros=macro_overrides
        )

        # don't write our new, weird manifest!
        self.linker = compile_manifest(self.config, self.manifest, write=False)
        return node

    def _raise_set_error(self):
        if self._raise_next_tick is not None:
            raise self._raise_next_tick

    def _in_thread(self, node, thread_done):
        runner = self.get_runner(node)
        try:
            self.node_results.append(runner.safe_run(self.manifest))
        except Exception as exc:
            self._raise_next_tick = exc
        finally:
            thread_done.set()

    def handle_request(self, name, sql, macros=None):
        # we could get a ctrl+c at any time, including during parsing.
        thread = None
        try:
            node = self._get_exec_node(name, sql, macros)

            selected_uids = [node.unique_id]
            self.runtime_cleanup(selected_uids)

            thread_done = threading.Event()
            thread = threading.Thread(target=self._in_thread,
                                      args=(node, thread_done))
            thread.start()
            thread_done.wait()
        except KeyboardInterrupt:
            adapter = get_adapter(self.config)
            if adapter.is_cancelable():

                for conn_name in adapter.cancel_open_connections():
                    rpc_logger.debug('canceled query {}'.format(conn_name))
                if thread:
                    thread.join()
            else:
                msg = ("The {} adapter does not support query "
                       "cancellation. Some queries may still be "
                       "running!".format(adapter.type()))

                rpc_logger.debug(msg)

            raise dbt.exceptions.RPCKilledException(signal.SIGINT)

        self._raise_set_error()
        return self.node_results[0].serialize()



from dbt.api import APIObject

from dbt.node_types import NodeType
from dbt.utils import deep_merge


UNPARSED_BASE_CONTRACT = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'package_name': {
            'type': 'string',
        },
        # filesystem
        'root_path': {
            'type': 'string',
            'description': 'The absolute path to the project root',
        },
        'path': {
            'type': 'string',
            'description': (
                'Relative path to the source file from the project root. '
                'Usually the same as original_file_path, but in some cases '
                'dbt will generate a path.'),
        },
        'original_file_path': {
            'type': 'string',
            'description': (
                'Relative path to the originating file from the project root.'
            ),
        }
    },
    'required': ['package_name', 'root_path', 'path', 'original_file_path']
}

UNPARSED_HAS_SQL_CONTRACT = {
    'properties': {
        'raw_sql': {
            'type': 'string',
            'description': (
                'For nodes defined in SQL files, this is just the contents '
                'of that file. For schema tests, snapshots, etc. this is '
                'generated by dbt.'),
        },
        'index': {
            'type': 'integer',
        }
    },
    'required': ['raw_sql']
}

UNPARSED_MACRO_CONTRACT = deep_merge(
    UNPARSED_BASE_CONTRACT,
    UNPARSED_HAS_SQL_CONTRACT
)

UNPARSED_NODE_CONTRACT = deep_merge(
    UNPARSED_BASE_CONTRACT,
    UNPARSED_HAS_SQL_CONTRACT,
    {
        'properties': {
            'name': {
                'type': 'string',
                'description': (
                    'Name of this node. For models, this is used as the '
                    'identifier in the database.'),
                'minLength': 1,
            },
            'resource_type': {
                'enum': [
                    NodeType.Model,
                    NodeType.Test,
                    NodeType.Analysis,
                    NodeType.Operation,
                    NodeType.Seed,
                    # we need this if parse_node is going to handle snapshots.
                    NodeType.Snapshot,
                    NodeType.RPCCall,
                ]
            },
        },
        'required': ['resource_type', 'name']
    }
)


class UnparsedMacro(APIObject):
    SCHEMA = UNPARSED_MACRO_CONTRACT


class UnparsedNode(APIObject):
    SCHEMA = UNPARSED_NODE_CONTRACT


COLUMN_TEST_CONTRACT = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'name': {
            'type': 'string',
            'description': 'The name of the column this test is for',
        },
        'description': {
            'type': 'string',
            'description': 'The description of this test',
        },
        'tests': {
            'type': 'array',
            'items': {
                "anyOf": [
                    # 'not_null', 'unique', ...
                    {'type': 'string'},
                    # 'relationships: {...}', 'accepted_values: {...}'
                    {'type': 'object', 'additionalProperties': True}
                ],
            },
            'description': 'The list of tests to perform',
        },
    },
    'required': ['name'],
}


UNPARSED_COLUMN_DESCRIPTION_CONTRACT = {
    'properties': {
        'columns': {
            'type': 'array',
            'items': COLUMN_TEST_CONTRACT,
        },
    },
}


UNPARSED_NODE_DESCRIPTION_CONTRACT = {
    'properties': {
        'name': {
            'type': 'string',
            'description': (
                'The name of this node, which is the name of the model it'
                'refers to'
            ),
            'minLength': 1,
        },
        'description': {
            'type': 'string',
            'description': (
                'The raw string description of the node after parsing the yaml'
            ),
        },
        'tests': {
            'type': 'array',
            'items': {
                "anyOf": [
                    {'type': 'string'},
                    {'type': 'object', 'additionalProperties': True}
                ],
            },
        },
    },
    'required': ['name'],
}


UNPARSED_NODE_UPDATE_CONTRACT = deep_merge(
    UNPARSED_NODE_DESCRIPTION_CONTRACT,
    UNPARSED_COLUMN_DESCRIPTION_CONTRACT,
    {
        'type': 'object',
        'additionalProperties': False,
        'description': (
            'A collection of the unparsed node updates, as provided in the '
            '"models" section of schema.yml'
        ),
    }
)


class UnparsedNodeUpdate(APIObject):
    """An unparsed node update is the blueprint for tests to be added and nodes
    to be updated, referencing a certain node (specifically, a Model or
    Source).
    """
    SCHEMA = UNPARSED_NODE_UPDATE_CONTRACT


TIME_CONTRACT = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'count': {
            'type': 'integer',
        },
        'period': {
            'enum': ['minute', 'hour', 'day'],
        },
    },
    'required': ['count', 'period'],
}


_FRESHNESS_CONTRACT = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'warn_after': {
            'anyOf': [
                {'type': 'null'},
                TIME_CONTRACT,
            ]
        },
        'error_after': {
            'anyOf': [
                {'type': 'null'},
                TIME_CONTRACT,

            ]
        },
    },
}


_QUOTING_CONTRACT = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'database': {'type': 'boolean'},
        'schema': {'type': 'boolean'},
        'identifier': {'type': 'boolean'},
    },
}


QUOTING_CONTRACT = {
    'properties': {
        'quoting': {
            'anyOf': [
                {'type': 'null'},
                _QUOTING_CONTRACT,
            ],
        },
    },
}


FRESHNESS_CONTRACT = {
    'properties': {
        'loaded_at_field': {
            'type': ['null', 'string'],
            'description': 'The field to use as the "loaded at" timestamp',
        },
        'freshness': {
            'anyOf': [
                {'type': 'null'},
                _FRESHNESS_CONTRACT,
            ],
        },
    },
}


UNPARSED_SOURCE_TABLE_DEFINITION_CONTRACT = deep_merge(
    UNPARSED_NODE_DESCRIPTION_CONTRACT,
    UNPARSED_COLUMN_DESCRIPTION_CONTRACT,
    FRESHNESS_CONTRACT,
    QUOTING_CONTRACT,
    {
        'description': (
            'A source table definition, as provided in the "tables" '
            'subsection of the "sources" section of schema.yml'
        ),
        'properties': {
            'identifier': {
                'type': 'string',
                'description': 'The identifier for the source table',
                'minLength': 1,
            },
        },
    }
)


UNPARSED_SOURCE_DEFINITION_CONTRACT = deep_merge(
    FRESHNESS_CONTRACT,
    QUOTING_CONTRACT,
    {
        'type': 'object',
        'additionalProperties': False,
        'description': (
            'A collection of the unparsed sources, as provided in the '
            '"sources" section of schema.yml'
        ),
        'properties': {
            'name': {
                'type': 'string',
                'description': 'The reference name of the source definition',
                'minLength': 1,
            },
            'loader': {
                'type': 'string',
                'description': 'The user-defined loader for this source',
                'minLength': 1,
            },
            'description': {
                'type': 'string',
                'description': 'The user-supplied description of the source',
            },
            'database': {
                'type': 'string',
                'description': 'The database name for the source table',
                'minLength': 1,
            },
            'schema': {
                'type': 'string',
                'description': 'The schema name for the source table',
                'minLength': 1,
            },
            'tables': {
                'type': 'array',
                'items': UNPARSED_SOURCE_TABLE_DEFINITION_CONTRACT,
                'description': 'The tables for this source',
                'minLength': 1,
            },
        },
        'required': ['name'],
    }
)


class UnparsedTableDefinition(APIObject):
    SCHEMA = UNPARSED_SOURCE_TABLE_DEFINITION_CONTRACT


class UnparsedSourceDefinition(APIObject):
    SCHEMA = UNPARSED_SOURCE_DEFINITION_CONTRACT

    @property
    def tables(self):
        return [UnparsedTableDefinition(**t) for t in self.get('tables', [])]


UNPARSED_DOCUMENTATION_FILE_CONTRACT = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'package_name': {
            'type': 'string',
        },
        # filesystem
        'root_path': {
            'type': 'string',
            'description': 'The absolute path to the project root',
        },
        'path': {
            'type': 'string',
            'description': (
                'Relative path to the source file from the project root. '
                'Usually the same as original_file_path, but in some cases '
                'dbt will generate a path.'),
        },
        'original_file_path': {
            'type': 'string',
            'description': (
                'Relative path to the originating file from the project root.'
            ),
        },
        'file_contents': {
            'type': 'string',
            'description': (
                'The raw text provided in the documentation block, presumably '
                'markdown.'
            ),
        },
        # TODO: I would like to remove this, but some graph error handling
        # cares about it.
        'resource_type': {
            'enum': [
                NodeType.Documentation,
            ]
        },
    },
    'required': [
        'package_name', 'root_path', 'path', 'original_file_path',
        'file_contents', 'resource_type'
    ],
}


class UnparsedDocumentationFile(APIObject):
    SCHEMA = UNPARSED_DOCUMENTATION_FILE_CONTRACT




import os

import dbt.contracts.project
import dbt.exceptions
import dbt.clients.system
import dbt.utils
import dbt.flags

from dbt.contracts.graph.unparsed import UnparsedNode
from dbt.parser.base import MacrosKnownParser
from dbt.node_types import NodeType


class BaseSqlParser(MacrosKnownParser):
    @classmethod
    def get_compiled_path(cls, name, relative_path):
        raise dbt.exceptions.NotImplementedException("Not implemented")

    def load_and_parse(self, package_name, root_dir, relative_dirs,
                       resource_type, tags=None):
        """Load and parse models in a list of directories. Returns a dict
           that maps unique ids onto ParsedNodes"""

        extension = "[!.#~]*.sql"

        if tags is None:
            tags = []

        if dbt.flags.STRICT_MODE:
            dbt.contracts.project.ProjectList(**self.all_projects)

        file_matches = dbt.clients.system.find_matching(
            root_dir,
            relative_dirs,
            extension)

        result = []

        for file_match in file_matches:
            file_contents = dbt.clients.system.load_file_contents(
                file_match.get('absolute_path'))

            parts = dbt.utils.split_path(file_match.get('relative_path', ''))
            name, _ = os.path.splitext(parts[-1])

            path = self.get_compiled_path(name,
                                          file_match.get('relative_path'))

            original_file_path = os.path.join(
                file_match.get('searched_path'),
                file_match.get('relative_path'))

            result.append({
                'name': name,
                'root_path': root_dir,
                'resource_type': resource_type,
                'path': path,
                'original_file_path': original_file_path,
                'package_name': package_name,
                'raw_sql': file_contents
            })

        return self.parse_sql_nodes(result, tags)

    def parse_sql_node(self, node_dict, tags=None):
        if tags is None:
            tags = []

        node = UnparsedNode(**node_dict)
        package_name = node.package_name

        unique_id = self.get_path(node.resource_type,
                                  package_name,
                                  node.name)

        project = self.all_projects.get(package_name)

        parse_ok = True
        if node.resource_type == NodeType.Model:
            parse_ok = self.check_block_parsing(
                node.name, node.original_file_path, node.raw_sql
            )

        node_parsed = self.parse_node(node, unique_id, project, tags=tags)
        if not parse_ok:
            # if we had a parse error in parse_node, we would not get here. So
            # this means we rejected a good file :(
            raise dbt.exceptions.InternalException(
                'the block parser rejected a good node: {} was marked invalid '
                'but is actually valid!'.format(node.original_file_path)
            )
        return unique_id, node_parsed

    def parse_sql_nodes(self, nodes, tags=None):
        if tags is None:
            tags = []

        results = SQLParseResult()

        for n in nodes:
            node_path, node_parsed = self.parse_sql_node(n, tags)

            # Ignore disabled nodes
            if not node_parsed.config['enabled']:
                results.disable(node_parsed)
                continue

            results.keep(node_path, node_parsed)

        return results


class SQLParseResult(object):
    def __init__(self):
        self.parsed = {}
        self.disabled = []

    def result(self, unique_id, node):
        if node.config['enabled']:
            self.keep(unique_id, node)
        else:
            self.disable(node)

    def disable(self, node):
        self.disabled.append(node)

    def keep(self, unique_id, node):
        if unique_id in self.parsed:
            dbt.exceptions.raise_duplicate_resource_name(
                self.parsed[unique_id], node
            )

        self.parsed[unique_id] = node

    def update(self, other):
        self.disabled.extend(other.disabled)
        for unique_id, node in other.parsed.items():
            self.keep(unique_id, node)



import dbt.exceptions

import dbt.context.common
from dbt.adapters.factory import get_adapter


def docs(unparsed, docrefs, column_name=None):

    def do_docs(*args):
        if len(args) != 1 and len(args) != 2:
            dbt.exceptions.doc_invalid_args(unparsed, args)
        doc_package_name = ''
        doc_name = args[0]
        if len(args) == 2:
            doc_package_name = args[1]

        docref = {
            'documentation_package': doc_package_name,
            'documentation_name': doc_name,
        }
        if column_name is not None:
            docref['column_name'] = column_name

        docrefs.append(docref)

        # IDK
        return True

    return do_docs


class Config(object):
    def __init__(self, model, source_config):
        self.model = model
        self.source_config = source_config

    def _transform_config(self, config):
        for oldkey in ('pre_hook', 'post_hook'):
            if oldkey in config:
                newkey = oldkey.replace('_', '-')
                if newkey in config:
                    dbt.exceptions.raise_compiler_error(
                        'Invalid config, has conflicting keys "{}" and "{}"'
                        .format(oldkey, newkey),
                        self.model
                    )
                config[newkey] = config.pop(oldkey)
        return config

    def __call__(self, *args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0:
            opts = args[0]
        elif len(args) == 0 and len(kwargs) > 0:
            opts = kwargs
        else:
            dbt.exceptions.raise_compiler_error(
                "Invalid inline model config",
                self.model)

        opts = self._transform_config(opts)

        self.source_config.update_in_model_config(opts)
        return ''

    def set(self, name, value):
        return self.__call__({name: value})

    def require(self, name, validator=None):
        return ''

    def get(self, name, validator=None, default=None):
        return ''


class DatabaseWrapper(dbt.context.common.BaseDatabaseWrapper):
    """The parser subclass of the database wrapper applies any explicit
    parse-time overrides.
    """
    def __getattr__(self, name):
        override = (name in self.adapter._available_ and
                    name in self.adapter._parse_replacements_)

        if override:
            return self.adapter._parse_replacements_[name]
        elif name in self.adapter._available_:
            return getattr(self.adapter, name)
        else:
            raise AttributeError(
                "'{}' object has no attribute '{}'".format(
                    self.__class__.__name__, name
                )
            )


class Var(dbt.context.common.Var):
    def get_missing_var(self, var_name):
        # in the parser, just always return None.
        return None


class RefResolver(dbt.context.common.BaseResolver):
    def __call__(self, *args):
        # When you call ref(), this is what happens at parse time
        if len(args) == 1 or len(args) == 2:
            self.model.refs.append(list(args))

        else:
            dbt.exceptions.ref_invalid_args(self.model, args)

        return self.Relation.create_from_node(self.config, self.model)


class SourceResolver(dbt.context.common.BaseResolver):
    def __call__(self, source_name, table_name):
        # When you call source(), this is what happens at parse time
        self.model.sources.append([source_name, table_name])
        return self.Relation.create_from_node(self.config, self.model)


class Provider(object):
    execute = False
    Config = Config
    DatabaseWrapper = DatabaseWrapper
    Var = Var
    ref = RefResolver
    source = SourceResolver


def generate(model, runtime_config, manifest, source_config):
    # during parsing, we don't have a connection, but we might need one, so we
    # have to acquire it.
    # In the future, it would be nice to lazily open the connection, as in some
    # projects it would be possible to parse without connecting to the db
    with get_adapter(runtime_config).connection_named(model.get('name')):
        return dbt.context.common.generate(
            model, runtime_config, manifest, source_config, Provider()
        )


def generate_macro(model, runtime_config, manifest):
    # parser.generate_macro is called by the get_${attr}_func family of Parser
    # methods, which preparse and cache the generate_${attr}_name family of
    # macros for use during parsing
    return dbt.context.common.generate_execute_macro(
        model, runtime_config, manifest, Provider()
    )



from dbt.api import APIObject
from dbt.utils import filter_null_values
from dbt.node_types import NodeType

import dbt.exceptions


class BaseRelation(APIObject):

    Table = "table"
    View = "view"
    CTE = "cte"
    MaterializedView = "materializedview"

    RelationTypes = [
        Table,
        View,
        CTE,
        MaterializedView
    ]

    DEFAULTS = {
        'metadata': {
            'type': 'BaseRelation'
        },
        'quote_character': '"',
        'quote_policy': {
            'database': True,
            'schema': True,
            'identifier': True
        },
        'include_policy': {
            'database': True,
            'schema': True,
            'identifier': True
        },
    }

    PATH_SCHEMA = {
        'type': 'object',
        'properties': {
            'database': {'type': ['string', 'null']},
            'schema': {'type': ['string', 'null']},
            'identifier': {'type': ['string', 'null']},
        },
        'required': ['database', 'schema', 'identifier'],
    }

    POLICY_SCHEMA = {
        'type': 'object',
        'properties': {
            'database': {'type': 'boolean'},
            'schema': {'type': 'boolean'},
            'identifier': {'type': 'boolean'},
        },
        'required': ['database', 'schema', 'identifier'],
    }

    SCHEMA = {
        'type': 'object',
        'properties': {
            'metadata': {
                'type': 'object',
                'properties': {
                    'type': {
                        'type': 'string',
                        'const': 'BaseRelation',
                    },
                },
            },
            'type': {
                'enum': RelationTypes + [None],
            },
            'path': PATH_SCHEMA,
            'include_policy': POLICY_SCHEMA,
            'quote_policy': POLICY_SCHEMA,
            'quote_character': {'type': 'string'},
        },
        'required': ['metadata', 'type', 'path', 'include_policy',
                     'quote_policy', 'quote_character']
    }

    PATH_ELEMENTS = ['database', 'schema', 'identifier']

    def matches(self, database=None, schema=None, identifier=None):
        search = filter_null_values({
            'database': database,
            'schema': schema,
            'identifier': identifier
        })

        if not search:
            # nothing was passed in
            raise dbt.exceptions.RuntimeException(
                "Tried to match relation, but no search path was passed!")

        exact_match = True
        approximate_match = True

        for k, v in search.items():
            if self.get_path_part(k) != v:
                exact_match = False

            if self.get_path_part(k).lower() != v.lower():
                approximate_match = False

        if approximate_match and not exact_match:
            target = self.create(
                database=database, schema=schema, identifier=identifier)
            dbt.exceptions.approximate_relation_match(target, self)

        return exact_match

    def get_path_part(self, part):
        return self.path.get(part)

    def should_quote(self, part):
        return self.quote_policy.get(part)

    def should_include(self, part):
        return self.include_policy.get(part)

    def quote(self, database=None, schema=None, identifier=None):
        policy = filter_null_values({
            'database': database,
            'schema': schema,
            'identifier': identifier
        })

        return self.incorporate(quote_policy=policy)

    def include(self, database=None, schema=None, identifier=None):
        policy = filter_null_values({
            'database': database,
            'schema': schema,
            'identifier': identifier
        })

        return self.incorporate(include_policy=policy)

    def information_schema(self, identifier=None):
        include_db = self.database is not None
        include_policy = filter_null_values({
            'database': include_db,
            'schema': True,
            'identifier': identifier is not None
        })
        quote_policy = filter_null_values({
            'database': self.quote_policy['database'],
            'schema': False,
            'identifier': False,
        })

        path_update = {
            'schema': 'information_schema',
            'identifier': identifier
        }

        return self.incorporate(
            quote_policy=quote_policy,
            include_policy=include_policy,
            path=path_update,
            table_name=identifier)

    def information_schema_only(self):
        return self.information_schema()

    def information_schema_table(self, identifier):
        return self.information_schema(identifier)

    def render(self, use_table_name=True):
        parts = []

        for k in self.PATH_ELEMENTS:
            if self.should_include(k):
                path_part = self.get_path_part(k)

                if path_part is None:
                    continue
                elif k == 'identifier':
                    if use_table_name:
                        path_part = self.table
                    else:
                        path_part = self.identifier

                parts.append(
                    self.quote_if(
                        path_part,
                        self.should_quote(k)))

        if len(parts) == 0:
            raise dbt.exceptions.RuntimeException(
                "No path parts are included! Nothing to render.")

        return '.'.join(parts)

    def quote_if(self, identifier, should_quote):
        if should_quote:
            return self.quoted(identifier)

        return identifier

    def quoted(self, identifier):
        return '{quote_char}{identifier}{quote_char}'.format(
            quote_char=self.quote_character,
            identifier=identifier)

    @classmethod
    def create_from_source(cls, source, **kwargs):
        quote_policy = dbt.utils.deep_merge(
            cls.DEFAULTS['quote_policy'],
            source.quoting,
            kwargs.get('quote_policy', {})
        )
        return cls.create(
            database=source.database,
            schema=source.schema,
            identifier=source.identifier,
            quote_policy=quote_policy,
            **kwargs
        )

    @classmethod
    def create_from_node(cls, config, node, table_name=None, quote_policy=None,
                         **kwargs):
        if quote_policy is None:
            quote_policy = {}

        quote_policy = dbt.utils.merge(config.quoting, quote_policy)

        return cls.create(
            database=node.get('database'),
            schema=node.get('schema'),
            identifier=node.get('alias'),
            table_name=table_name,
            quote_policy=quote_policy,
            **kwargs)

    @classmethod
    def create_from(cls, config, node, **kwargs):
        if node.resource_type == NodeType.Source:
            return cls.create_from_source(node, **kwargs)
        else:
            return cls.create_from_node(config, node, **kwargs)

    @classmethod
    def create(cls, database=None, schema=None,
               identifier=None, table_name=None,
               type=None, **kwargs):
        if table_name is None:
            table_name = identifier

        return cls(type=type,
                   path={
                       'database': database,
                       'schema': schema,
                       'identifier': identifier
                   },
                   table_name=table_name,
                   **kwargs)

    def __repr__(self):
        return "<{} {}>".format(self.__class__.__name__, self.render())

    def __hash__(self):
        return hash(self.render())

    def __str__(self):
        return self.render()

    @property
    def path(self):
        return self.get('path', {})

    @property
    def database(self):
        return self.path.get('database')

    @property
    def schema(self):
        return self.path.get('schema')

    @property
    def identifier(self):
        return self.path.get('identifier')

    # Here for compatibility with old Relation interface
    @property
    def name(self):
        return self.identifier

    # Here for compatibility with old Relation interface
    @property
    def table(self):
        return self.table_name

    @property
    def is_table(self):
        return self.type == self.Table

    @property
    def is_cte(self):
        return self.type == self.CTE

    @property
    def is_view(self):
        return self.type == self.View


class Column(object):
    TYPE_LABELS = {
        'STRING': 'TEXT',
        'TIMESTAMP': 'TIMESTAMP',
        'FLOAT': 'FLOAT',
        'INTEGER': 'INT'
    }

    def __init__(self, column, dtype, char_size=None, numeric_precision=None,
                 numeric_scale=None):
        self.column = column
        self.dtype = dtype
        self.char_size = char_size
        self.numeric_precision = numeric_precision
        self.numeric_scale = numeric_scale

    @classmethod
    def translate_type(cls, dtype):
        return cls.TYPE_LABELS.get(dtype.upper(), dtype)

    @classmethod
    def create(cls, name, label_or_dtype):
        column_type = cls.translate_type(label_or_dtype)
        return cls(name, column_type)

    @property
    def name(self):
        return self.column

    @property
    def quoted(self):
        return '"{}"'.format(self.column)

    @property
    def data_type(self):
        if self.is_string():
            return Column.string_type(self.string_size())
        elif self.is_numeric():
            return Column.numeric_type(self.dtype, self.numeric_precision,
                                       self.numeric_scale)
        else:
            return self.dtype

    def is_string(self):
        return self.dtype.lower() in ['text', 'character varying', 'character',
                                      'varchar']

    def is_numeric(self):
        return self.dtype.lower() in ['numeric', 'number']

    def string_size(self):
        if not self.is_string():
            raise RuntimeError("Called string_size() on non-string field!")

        if self.dtype == 'text' or self.char_size is None:
            # char_size should never be None. Handle it reasonably just in case
            return 256
        else:
            return int(self.char_size)

    def can_expand_to(self, other_column):
        """returns True if this column can be expanded to the size of the
        other column"""
        if not self.is_string() or not other_column.is_string():
            return False

        return other_column.string_size() > self.string_size()

    def literal(self, value):
        return "{}::{}".format(value, self.data_type)

    @classmethod
    def string_type(cls, size):
        return "character varying({})".format(size)

    @classmethod
    def numeric_type(cls, dtype, precision, scale):
        # This could be decimal(...), numeric(...), number(...)
        # Just use whatever was fed in here -- don't try to get too clever
        if precision is None or scale is None:
            return dtype
        else:
            return "{}({},{})".format(dtype, precision, scale)

    def __repr__(self):
        return "<Column {} ({})>".format(self.name, self.data_type)



from dbt.utils import get_materialization, add_ephemeral_model_prefix

import dbt.clients.jinja
import dbt.context.common
import dbt.flags
from dbt.parser import ParserUtils

from dbt.logger import GLOBAL_LOGGER as logger  # noqa


class BaseRefResolver(dbt.context.common.BaseResolver):
    def resolve(self, args):
        name = None
        package = None

        if len(args) == 1:
            name = args[0]
        elif len(args) == 2:
            package, name = args
        else:
            dbt.exceptions.ref_invalid_args(self.model, args)

        target_model = ParserUtils.resolve_ref(
            self.manifest,
            name,
            package,
            self.current_project,
            self.model.package_name)

        if target_model is None or target_model is ParserUtils.DISABLED:
            dbt.exceptions.ref_target_not_found(
                self.model,
                name,
                package)
        return target_model, name

    def create_ephemeral_relation(self, target_model, name):
        self.model.set_cte(target_model.unique_id, None)
        return self.Relation.create(
            type=self.Relation.CTE,
            identifier=add_ephemeral_model_prefix(name)
        ).quote(identifier=False)

    def create_relation(self, target_model, name):
        if get_materialization(target_model) == 'ephemeral':
            return self.create_ephemeral_relation(target_model, name)
        else:
            return self.Relation.create_from_node(self.config, target_model)


class RefResolver(BaseRefResolver):
    def validate(self, resolved, args):
        if resolved.unique_id not in self.model.depends_on.get('nodes'):
            dbt.exceptions.ref_bad_context(self.model, args)

    def __call__(self, *args):
        # When you call ref(), this is what happens at runtime
        target_model, name = self.resolve(args)
        self.validate(target_model, args)
        return self.create_relation(target_model, name)


class SourceResolver(dbt.context.common.BaseResolver):
    def resolve(self, source_name, table_name):
        target_source = ParserUtils.resolve_source(
            self.manifest,
            source_name,
            table_name,
            self.current_project,
            self.model.package_name
        )

        if target_source is None:
            dbt.exceptions.source_target_not_found(
                self.model,
                source_name,
                table_name)
        return target_source

    def __call__(self, source_name, table_name):
        """When you call source(), this is what happens at runtime"""
        target_source = self.resolve(source_name, table_name)
        return self.Relation.create_from_source(target_source)


class Config:
    def __init__(self, model, source_config=None):
        self.model = model
        # we never use or get a source config, only the parser cares

    def __call__(*args, **kwargs):
        return ''

    def set(self, name, value):
        return self.__call__({name: value})

    def _validate(self, validator, value):
        validator(value)

    def require(self, name, validator=None):
        if name not in self.model['config']:
            dbt.exceptions.missing_config(self.model, name)

        to_return = self.model['config'][name]

        if validator is not None:
            self._validate(validator, to_return)

        return to_return

    def get(self, name, validator=None, default=None):
        to_return = self.model['config'].get(name, default)

        if validator is not None and default is not None:
            self._validate(validator, to_return)

        return to_return


class DatabaseWrapper(dbt.context.common.BaseDatabaseWrapper):
    """The runtime database wrapper exposes everything the adapter marks
    available.
    """
    def __getattr__(self, name):
        if name in self.adapter._available_:
            return getattr(self.adapter, name)
        else:
            raise AttributeError(
                "'{}' object has no attribute '{}'".format(
                    self.__class__.__name__, name
                )
            )


class Var(dbt.context.common.Var):
    pass


class Provider(object):
    execute = True
    Config = Config
    DatabaseWrapper = DatabaseWrapper
    Var = Var
    ref = RefResolver
    source = SourceResolver


def generate(model, runtime_config, manifest):
    return dbt.context.common.generate(
        model, runtime_config, manifest, None, Provider())



from jsonrpc.exceptions import \
    JSONRPCDispatchException, \
    JSONRPCInvalidParams, \
    JSONRPCParseError, \
    JSONRPCInvalidRequestException, \
    JSONRPCInvalidRequest
from jsonrpc import JSONRPCResponseManager
from jsonrpc.jsonrpc import JSONRPCRequest
from jsonrpc.jsonrpc2 import JSONRPC20Response

import json
import uuid
import multiprocessing
import os
import signal
import time
from collections import namedtuple

from dbt.adapters.factory import load_plugin
from dbt.compat import QueueEmpty
from dbt import flags
from dbt.logger import RPC_LOGGER as logger
from dbt.logger import add_queue_handler
import dbt.exceptions


class RPCException(JSONRPCDispatchException):
    def __init__(self, code=None, message=None, data=None, logs=None):
        if code is None:
            code = -32000
        if message is None:
            message = 'Server error'
        if data is None:
            data = {}

        super(RPCException, self).__init__(code=code, message=message,
                                           data=data)
        self.logs = logs

    def __str__(self):
        return (
            'RPCException({0.code}, {0.message}, {0.data}, {1.logs})'
            .format(self.error, self)
        )

    @property
    def logs(self):
        return self.error.data.get('logs')

    @logs.setter
    def logs(self, value):
        if value is None:
            return
        self.error.data['logs'] = value

    @classmethod
    def from_error(cls, err):
        return cls(err.code, err.message, err.data, err.data.get('logs'))


def invalid_params(data):
    return RPCException(
        code=JSONRPCInvalidParams.CODE,
        message=JSONRPCInvalidParams.MESSAGE,
        data=data
    )


def server_error(err, logs=None):
    exc = dbt.exceptions.Exception(str(err))
    return dbt_error(exc, logs)


def timeout_error(timeout_value, logs=None):
    exc = dbt.exceptions.RPCTimeoutException(timeout_value)
    return dbt_error(exc, logs)


def dbt_error(exc, logs=None):
    exc = RPCException(code=exc.CODE, message=exc.MESSAGE, data=exc.data(),
                       logs=logs)
    return exc


class QueueMessageType(object):
    Error = 'error'
    Result = 'result'
    Log = 'log'

    @classmethod
    def terminating(cls):
        return [
            cls.Error,
            cls.Result
        ]


def sigterm_handler(signum, frame):
    raise dbt.exceptions.RPCKilledException(signum)


class RequestDispatcher(object):
    """A special dispatcher that knows about requests."""
    def __init__(self, http_request, json_rpc_request, manager):
        self.http_request = http_request
        self.json_rpc_request = json_rpc_request
        self.manager = manager
        self.task = None

    def rpc_factory(self, task):
        request_handler = RequestTaskHandler(task,
                                             self.http_request,
                                             self.json_rpc_request)

        def rpc_func(**kwargs):
            try:
                self.manager.add_request(request_handler)
                return request_handler.handle(kwargs)
            finally:
                self.manager.mark_done(request_handler)

        return rpc_func

    def __getitem__(self, key):
        # the dispatcher's keys are method names and its values are functions
        # that implement the RPC calls
        func = self.manager.rpc_builtin(key)
        if func is not None:
            return func

        task = self.manager.rpc_task(key)
        return self.rpc_factory(task)


def _nt_setup(config, args):
    """
    On windows, we have to do a some things that dbt does dynamically at
    process load.

    These things are inherited automatically on posix, where fork() keeps
    everything in memory.
    """
    # reset flags
    flags.set_from_args(args)
    # reload the active plugin
    load_plugin(config.credentials.type)

    # reset tracking, etc
    config.config.set_values(args.profiles_dir)


def _task_bootstrap(task, queue, kwargs):
    signal.signal(signal.SIGTERM, sigterm_handler)
    # the first thing we do in a new process: start logging
    add_queue_handler(queue)
    # on windows, we need to reload our plugins because of how it starts new
    # processes. At this point there are no adapter plugins loaded!
    if os.name == 'nt':
        _nt_setup(task.config, task.args)

    error = None
    result = None
    try:
        result = task.handle_request(**kwargs)
    except RPCException as exc:
        error = exc
    except dbt.exceptions.RPCKilledException as exc:
        # do NOT log anything here, you risk triggering a deadlock on the
        # queue handler we inserted above
        error = dbt_error(exc)
    except dbt.exceptions.Exception as exc:
        logger.debug('dbt runtime exception', exc_info=True)
        error = dbt_error(exc)
    except Exception as exc:
        logger.debug('uncaught python exception', exc_info=True)
        error = server_error(exc)

    # put whatever result we got onto the queue as well.
    if error is not None:
        queue.put([QueueMessageType.Error, error.error])
    else:
        queue.put([QueueMessageType.Result, result])


class RequestTaskHandler(object):
    def __init__(self, task, http_request, json_rpc_request):
        self.task = task
        self.http_request = http_request
        self.json_rpc_request = json_rpc_request
        self.queue = None
        self.process = None
        self.started = None
        self.timeout = None
        self.logs = []
        self.task_id = uuid.uuid4()

    @property
    def request_source(self):
        return self.http_request.remote_addr

    @property
    def request_id(self):
        return self.json_rpc_request._id

    @property
    def method(self):
        return self.task.METHOD_NAME

    def _next_timeout(self):
        if self.timeout is None:
            return None
        end = self.started + self.timeout
        timeout = end - time.time()
        if timeout < 0:
            raise dbt.exceptions.RPCTimeoutException(self.timeout)
        return timeout

    def _wait_for_results(self):
        """Wait for results off the queue. If there is a timeout set, and it is
        exceeded, raise an RPCTimeoutException.
        """
        while True:
            get_timeout = self._next_timeout()
            try:
                msgtype, value = self.queue.get(timeout=get_timeout)
            except QueueEmpty:
                raise dbt.exceptions.RPCTimeoutException(self.timeout)

            if msgtype == QueueMessageType.Log:
                self.logs.append(value)
            elif msgtype in QueueMessageType.terminating():
                return msgtype, value
            else:
                raise dbt.exceptions.InternalException(
                    'Got invalid queue message type {}'.format(msgtype)
                )

    def _join_process(self):
        try:
            msgtype, result = self._wait_for_results()
        except dbt.exceptions.RPCTimeoutException:
            self.process.terminate()
            raise timeout_error(self.timeout)
        except dbt.exceptions.Exception as exc:
            raise dbt_error(exc)
        except Exception as exc:
            raise server_error(exc)
        finally:
            self.process.join()

        if msgtype == QueueMessageType.Error:
            raise RPCException.from_error(result)

        return result

    def get_result(self):
        try:
            result = self._join_process()
        except RPCException as exc:
            exc.logs = self.logs
            raise

        result['logs'] = self.logs
        return result

    def handle(self, kwargs):
        self.started = time.time()
        self.timeout = kwargs.pop('timeout', None)
        self.queue = multiprocessing.Queue()
        self.process = multiprocessing.Process(
            target=_task_bootstrap,
            args=(self.task, self.queue, kwargs)
        )
        self.process.start()
        result = self.get_result()
        return result

    @property
    def state(self):
        if self.started is None:
            return 'not started'
        elif self.process is None:
            return 'initializing'
        elif self.process.is_alive():
            return 'running'
        else:
            return 'finished'


TaskRow = namedtuple(
    'TaskRow',
    'task_id request_id request_source method state start elapsed timeout'
)


class TaskManager(object):
    def __init__(self):
        self.tasks = {}
        self.completed = {}
        self._rpc_task_map = {}
        self._rpc_function_map = {}
        self._lock = multiprocessing.Lock()

    def add_request(self, request_handler):
        self.tasks[request_handler.task_id] = request_handler

    def add_task_handler(self, task):
        self._rpc_task_map[task.METHOD_NAME] = task

    def rpc_task(self, method_name):
        return self._rpc_task_map[method_name]

    def process_listing(self, active=True, completed=False):
        included_tasks = {}
        with self._lock:
            if completed:
                included_tasks.update(self.completed)
            if active:
                included_tasks.update(self.tasks)

        table = []
        now = time.time()
        for task_handler in included_tasks.values():
            start = task_handler.started
            if start is not None:
                elapsed = now - start

            table.append(TaskRow(
                str(task_handler.task_id), task_handler.request_id,
                task_handler.request_source, task_handler.method,
                task_handler.state, start, elapsed, task_handler.timeout
            ))
        table.sort(key=lambda r: (r.state, r.start))
        result = {
            'rows': [dict(r._asdict()) for r in table],
        }
        return result

    def process_kill(self, task_id):
        # TODO: this result design is terrible
        result = {
            'found': False,
            'started': False,
            'finished': False,
            'killed': False
        }
        task_id = uuid.UUID(task_id)
        try:
            task = self.tasks[task_id]
        except KeyError:
            # nothing to do!
            return result

        result['found'] = True

        if task.process is None:
            return result
        pid = task.process.pid
        if pid is None:
            return result

        result['started'] = True

        if task.process.is_alive():
            os.kill(pid, signal.SIGINT)
            result['killed'] = True
            return result

        result['finished'] = True
        return result

    def rpc_builtin(self, method_name):
        if method_name == 'ps':
            return self.process_listing
        if method_name == 'kill' and os.name != 'nt':
            return self.process_kill
        return None

    def mark_done(self, request_handler):
        task_id = request_handler.task_id
        with self._lock:
            if task_id not in self.tasks:
                # lost a task! Maybe it was killed before it started.
                return
            self.completed[task_id] = self.tasks.pop(task_id)

    def methods(self):
        rpc_builtin_methods = ['ps']
        if os.name != 'nt':
            rpc_builtin_methods.append('kill')
        return list(self._rpc_task_map) + rpc_builtin_methods


class ResponseManager(JSONRPCResponseManager):
    """Override the default response manager to handle request metadata and
    track in-flight tasks.
    """
    @classmethod
    def handle(cls, http_request, task_manager):
        # pretty much just copy+pasted from the original, with slight tweaks to
        # preserve the request
        request_str = http_request.data
        if isinstance(request_str, bytes):
            request_str = request_str.decode("utf-8")

        try:
            data = json.loads(request_str)
        except (TypeError, ValueError):
            return JSONRPC20Response(error=JSONRPCParseError()._data)

        try:
            request = JSONRPCRequest.from_data(data)
        except JSONRPCInvalidRequestException:
            return JSONRPC20Response(error=JSONRPCInvalidRequest()._data)

        dispatcher = RequestDispatcher(
            http_request,
            request,
            task_manager
        )

        return cls.handle_request(request, dispatcher)



import re

import dbt.exceptions


def regex(pat):
    return re.compile(pat, re.DOTALL | re.MULTILINE)


class BlockData(object):
    """raw plaintext data from the top level of the file."""
    def __init__(self, contents):
        self.block_type_name = '__dbt__data'
        self.contents = contents
        self.full_block = contents


class BlockTag(object):
    def __init__(self, block_type_name, block_name, contents=None,
                 full_block=None, **kw):
        self.block_type_name = block_type_name
        self.block_name = block_name
        self.contents = contents
        self.full_block = full_block

    def __str__(self):
        return 'BlockTag({!r}, {!r})'.format(self.block_type_name,
                                             self.block_name)

    def __repr__(self):
        return str(self)

    @property
    def end_block_type_name(self):
        return 'end{}'.format(self.block_type_name)

    def end_pat(self):
        # we don't want to use string formatting here because jinja uses most
        # of the string formatting operators in its syntax...
        pattern = ''.join((
            r'(?P<endblock>((?:\s*\{\%\-|\{\%)\s*',
            self.end_block_type_name,
            r'\s*(?:\-\%\}\s*|\%\})))',
        ))
        return regex(pattern)


_NAME_PATTERN = r'[A-Za-z_][A-Za-z_0-9]*'

COMMENT_START_PATTERN = regex(r'(?:(?P<comment_start>(\s*\{\#)))')
COMMENT_END_PATTERN = regex(r'(.*?)(\s*\#\})')
RAW_START_PATTERN = regex(
    r'(?:\s*\{\%\-|\{\%)\s*(?P<raw_start>(raw))\s*(?:\-\%\}\s*|\%\})'
)
EXPR_START_PATTERN = regex(r'(?P<expr_start>(\{\{\s*))')
EXPR_END_PATTERN = regex(r'(?P<expr_end>(\s*\}\}))')

BLOCK_START_PATTERN = regex(''.join((
    r'(?:\s*\{\%\-|\{\%)\s*',
    r'(?P<block_type_name>({}))'.format(_NAME_PATTERN),
    # some blocks have a 'block name'.
    r'(?:\s+(?P<block_name>({})))?'.format(_NAME_PATTERN),
)))

TAG_CLOSE_PATTERN = regex(r'(?:(?P<tag_close>(\-\%\}\s*|\%\})))')
# if you do {% materialization foo, adapter="myadapter' %} and end up with
# mismatched quotes this will still match, but jinja will fail somewhere
# since the adapter= argument has to be an adapter name, and none have quotes
# or anything else in them. So this should be fine.
MATERIALIZATION_ARGS_PATTERN = regex(
    r'\s*,\s*'
    r'''(?P<adpater_arg>(adapter=(?:['"]{}['"])|default))'''
    .format(_NAME_PATTERN)
)
# macros an stuff like macros get open parents, followed by a very complicated
# argument spec! In fact, it's easiest to parse it in tiny little chunks
# because we have to handle awful stuff like string parsing ;_;
MACRO_ARGS_START_PATTERN = regex(r'\s*(?P<macro_start>\()\s*')
MACRO_ARGS_END_PATTERN = regex(r'\s*(?P<macro_end>(\)))\s*')

# macros can be like {% macro foo(bar) %} or {% macro foo(bar, baz) %} or
# {% macro foo(bar, baz="quux") %} or ...
# I think jinja disallows default values after required (like Python), but we
# can ignore that and let jinja deal
MACRO_ARG_PATTERN = regex(''.join((
    r'\s*(?P<macro_arg_name>({}))\s*',
    r'((?P<value>=)|(?P<more_args>,)?)\s*'.format(_NAME_PATTERN),
)))

# stolen from jinja's lexer. Note that we've consumed all prefix whitespace by
# the time we want to use this.
STRING_PATTERN = regex(
    r"(?P<string>('([^'\\]*(?:\\.[^'\\]*)*)'|"
    r'"([^"\\]*(?:\\.[^"\\]*)*)"))'
)

QUOTE_START_PATTERN = regex(r'''(?P<quote>(['"]))''')

# any number of non-quote characters, followed by:
# - quote: a quote mark indicating start of a string (you'll want to backtrack
#          the regex end on quotes and then match with the string pattern)
# - a comma (so there will be another full argument)
# - a closing parenthesis (you can now expect a closing tag)
NON_STRING_MACRO_ARGS_PATTERN = regex(
    # anything, followed by a quote, open/close paren, or comma
    r'''(.*?)'''
    r'''((?P<quote>(['"]))|(?P<open>(\())|(?P<close>(\)))|(?P<comma>(\,)))'''
)


NON_STRING_DO_BLOCK_MEMBER_PATTERN = regex(
    # anything, followed by a quote, paren, or a tag end
    r'''(.*?)'''
    r'''((?P<quote>(['"]))|(?P<open>(\())|(?P<close>(\))))'''
)


class BlockIterator(object):
    def __init__(self, data):
        self.data = data
        self.blocks = []
        self._block_contents = None
        self._parenthesis_stack = []
        self.pos = 0

    def advance(self, new_position):
        blk = self.data[self.pos:new_position]

        if self._block_contents is not None:
            self._block_contents += blk

        self.pos = new_position

    def rewind(self, amount=1):
        if self._block_contents is not None:
            self._block_contents = self._block_contents[:-amount]

        self.pos -= amount

    def _search(self, pattern):
        return pattern.search(self.data, self.pos)

    def _match(self, pattern):
        return pattern.match(self.data, self.pos)

    def expect_comment_end(self):
        """Expect a comment end and return the match object.
        """
        match = self._expect_match('#}', COMMENT_END_PATTERN)
        self.advance(match.end())

    def expect_raw_end(self):
        end_pat = BlockTag('raw', None).end_pat()
        match = self._search(end_pat)
        if match is None:
            dbt.exceptions.raise_compiler_error(
                'unexpected EOF, expected {% endraw %}'
            )
        self.advance(match.end())

    def _first_match(self, *patterns, **kwargs):
        matches = []
        for pattern in patterns:
            # default to 'search', but sometimes we want to 'match'.
            if kwargs.get('method', 'search') == 'search':
                match = self._search(pattern)
            else:
                match = self._match(pattern)
            if match:
                matches.append(match)
        if not matches:
            return None
        # if there are multiple matches, pick the least greedy match
        # TODO: do I need to account for m.start(), or is this ok?
        return min(matches, key=lambda m: m.end())

    def _expect_match(self, expected_name, *patterns, **kwargs):
        match = self._first_match(*patterns, **kwargs)
        if match is None:
            msg = 'unexpected EOF, expected {}, got "{}"'.format(
                expected_name, self.data[self.pos:]
            )
            dbt.exceptions.raise_compiler_error(msg)
        return match

    def handle_expr(self):
        """Handle an expression. At this point we're at a string like:
            {{ 1 + 2 }}
               ^ right here

        We expect to find a `}}`, but we might find one in a string before
        that. Imagine the case of `{{ 2 * "}}" }}`...

        You're not allowed to have blocks or comments inside an expr so it is
        pretty straightforward, I hope: only strings can get in the way.
        """
        while True:
            match = self._expect_match('}}',
                                       EXPR_END_PATTERN,
                                       QUOTE_START_PATTERN)
            if match.groupdict().get('expr_end') is not None:
                break
            else:
                # it's a quote. we haven't advanced for this match yet, so
                # just slurp up the whole string, no need to rewind.
                match = self._expect_match('string', STRING_PATTERN)
                self.advance(match.end())

        self.advance(match.end())

    def handle_block(self, match, block_start=None):
        """Handle a block. The current state of the parser should be after the
        open block is completed:
            {% blk foo %}my data {% endblk %}
                         ^ right here
        """
        # we have to handle comments inside blocks because you could do this:
        # {% blk foo %}asdf {# {% endblk %} #} {%endblk%}
        # they still end up in the data/raw_data of the block itself, but we
        # have to know to ignore stuff until the end comment marker!
        found = BlockTag(**match.groupdict())
        # the full block started at the given match start, which may include
        # prefixed whitespace! we'll strip it later
        if block_start is None:
            block_start = match.start()

        self._block_contents = ''

        search = [found.end_pat(), COMMENT_START_PATTERN, RAW_START_PATTERN,
                  EXPR_START_PATTERN]

        # docs and macros do not honor embedded quotes
        if found.block_type_name not in ('docs', 'macro'):
            # is this right?
            search.append(QUOTE_START_PATTERN)

        # you can have as many comments in your block as you'd like!
        while True:
            match = self._expect_match(
                '"{}"'.format(found.end_block_type_name), *search
            )
            groups = match.groupdict()
            if groups.get('endblock') is not None:
                break

            self.advance(match.end())

            if groups.get('comment_start') is not None:
                self.expect_comment_end()
            elif groups.get('raw_start') is not None:
                self.expect_raw_end()
            elif groups.get('quote') is not None:
                self.rewind()
                match = self._expect_match('any string', STRING_PATTERN)
                self.advance(match.end())
            elif groups.get('expr_start') is not None:
                self.handle_expr()
            else:
                raise dbt.exceptions.InternalException(
                    'unhandled regex in handle_block, no match: {}'
                    .format(groups)
                )

        # we want to advance to just the end tag at first, to extract the
        # contents
        self.advance(match.start())
        found.contents = self._block_contents
        self._block_contents = None
        # now advance to the end
        self.advance(match.end())
        found.full_block = self.data[block_start:self.pos]
        return found

    def handle_materialization(self, match):
        self._expect_match('materialization args',
                           MATERIALIZATION_ARGS_PATTERN)
        endtag = self._expect_match('%}', TAG_CLOSE_PATTERN)
        self.advance(endtag.end())
        # handle the block we started with!
        self.blocks.append(self.handle_block(match))

    def handle_do(self, match, expect_block):
        if expect_block:
            # we might be wrong to expect a block ({% do (...) %}, for example)
            # so see if there's more data before the tag closes. if there
            # isn't, we expect a block.
            close_match = self._expect_match('%}', TAG_CLOSE_PATTERN)
            unprocessed = self.data[match.end():close_match.start()].strip()
            expect_block = not unprocessed

        if expect_block:
            # if we're here, expect_block is True and we must have set
            # close_match
            self.advance(close_match.end())
            block = self.handle_block(match)
        else:
            # we have a do-statement like {% do thing() %}, so no {% enddo %}
            # also, we don't want to advance to the end of the match, as it
            # might be inside a string or something! So go back and figure out
            self._process_rval_components()
            block = BlockTag('do', None,
                             full_block=self.data[match.start():self.pos])
        self.blocks.append(block)

    def handle_set(self, match):
        equal_or_close = self._expect_match('%} or =',
                                            TAG_CLOSE_PATTERN, regex(r'='))
        self.advance(equal_or_close.end())
        if equal_or_close.groupdict().get('tag_close') is None:
            # it's an equals sign, must be like {% set x = 1 %}
            self._process_rval_components()
            # watch out, order matters here on python 2
            block = BlockTag(full_block=self.data[match.start():self.pos],
                             **match.groupdict())
        else:
            # it's a tag close, must be like {% set x %}...{% endset %}
            block = self.handle_block(match)
        self.blocks.append(block)

    def find_block(self):
        open_block = (
            r'(?:\s*\{\%\-|\{\%)\s*'
            r'(?P<block_type_name>([A-Za-z_][A-Za-z_0-9]*))'
            # some blocks have a 'block name'.
            r'(?:\s+(?P<block_name>([A-Za-z_][A-Za-z_0-9]*)))?'
        )

        match = self._first_match(regex(open_block), COMMENT_START_PATTERN)
        if match is None:
            return False

        raw_toplevel = self.data[self.pos:match.start()]
        if len(raw_toplevel) > 0:
            self.blocks.append(BlockData(raw_toplevel))

        matchgroups = match.groupdict()

        # comments are easy
        if matchgroups.get('comment_start') is not None:
            start = match.start()
            self.advance(match.end())
            self.expect_comment_end()
            self.blocks.append(BlockData(self.data[start:self.pos]))
            return True

        block_type_name = matchgroups.get('block_type_name')

        if block_type_name == 'raw':
            start = match.start()
            self.expect_raw_end()
            self.blocks.append(BlockData(self.data[start:self.pos]))
            return True

        if block_type_name == 'materialization':
            self.advance(match.end())
            self.handle_materialization(match)
            return True

        if block_type_name == 'do':
            # if there is a "block_name" in the match groups, we don't expect a
            # block as the "block name" is actually part of the do-statement.
            # we need to do this to handle the (weird and probably wrong!) case
            # of a do-statement that is only a single identifier - techincally
            # allowed in jinja. (for example, {% do thing %})
            expect_block = matchgroups.get('block_name') is None
            self.handle_do(match, expect_block=expect_block)
            return True

        if block_type_name == 'set':
            self.advance(match.end())
            self.handle_set(match)
            return True

        # we're somewhere like this {% block_type_name block_type
        # we've either got arguments, a close of tag (%}), or bad input.
        # we've handled materializations already (they're weird!)
        # thankfully, comments aren't allowed *inside* a block def...
        block_end_match = self._expect_match('%} or (...)',
                                             TAG_CLOSE_PATTERN,
                                             MACRO_ARGS_START_PATTERN)
        self.advance(block_end_match.end())
        if block_end_match.groupdict().get('macro_start') is not None:
            # we've hit our first parenthesis!
            self._parenthesis_stack = [True]
            self._process_macro_args()
            self.advance(self._expect_match('%}', TAG_CLOSE_PATTERN).end())

        # tag close time!
        self.blocks.append(self.handle_block(match))
        return True

    def _process_rval_components(self):
        """This is suspiciously similar to _process_macro_default_arg, probably
        want to figure out how to merge the two.

        Process the rval of an assignment statement or a do-block
        """
        while True:
            match = self._expect_match(
                'do block component',
                # you could have a string, though that would be weird
                STRING_PATTERN,
                # a quote or an open/close parenthesis
                NON_STRING_DO_BLOCK_MEMBER_PATTERN,
                # a tag close
                TAG_CLOSE_PATTERN
            )
            matchgroups = match.groupdict()
            self.advance(match.end())
            if matchgroups.get('string') is not None:
                continue
            elif matchgroups.get('quote') is not None:
                self.rewind()
                # now look for a string
                match = self._expect_match('any string', STRING_PATTERN)
                self.advance(match.end())
            elif matchgroups.get('open'):
                self._parenthesis_stack.append(True)
            elif matchgroups.get('close'):
                self._parenthesis_stack.pop()
            elif matchgroups.get('tag_close'):
                if self._parenthesis_stack:
                    msg = ('Found "%}", expected ")"')
                    dbt.exceptions.raise_compiler_error(msg)
                return
            # else whitespace

    def _process_macro_default_arg(self):
        """Handle the bit after an '=' in a macro default argument. This is
        probably the trickiest thing. The goal here is to accept all strings
        jinja would accept and always handle block start/end correctly: It's
        fine to have false positives, jinja can fail later.

        Return True if there are more arguments expected.
        """
        while self._parenthesis_stack:
            match = self._expect_match(
                'macro argument',
                # you could have a string
                STRING_PATTERN,
                # a quote, a comma, or a open/close parenthesis
                NON_STRING_MACRO_ARGS_PATTERN,
                # we want to "match", not "search"
                method='match'
            )
            matchgroups = match.groupdict()
            self.advance(match.end())
            if matchgroups.get('string') is not None:
                # we got a string value. There could be more data.
                continue
            elif matchgroups.get('quote') is not None:
                # we got a bunch of data and then a string opening value.
                # put the quote back on the menu
                self.rewind()
                # now look for a string
                match = self._expect_match('any string', STRING_PATTERN)
                self.advance(match.end())
            elif matchgroups.get('comma') is not None:
                # small hack: if we hit a comma and there is one parenthesis
                # left, return to look for a new name. otherwise we're still
                # looking for the parameter close.
                if len(self._parenthesis_stack) == 1:
                    return
            elif matchgroups.get('open'):
                self._parenthesis_stack.append(True)
            elif matchgroups.get('close'):
                self._parenthesis_stack.pop()
            else:
                raise dbt.exceptions.InternalException(
                    'unhandled regex in _process_macro_default_arg(), no match'
                    ': {}'.format(matchgroups)
                )

    def _process_macro_args(self):
        """Macro args are pretty tricky! Arg names themselves are simple, but
        you can set arbitrary default values, including doing stuff like:
        {% macro my_macro(arg="x" + ("}% {# {% endmacro %}" * 2)) %}

        Which makes you a jerk, but is valid jinja.
        """
        # we are currently after the first parenthesis (+ any whitespace) after
        # the macro args started. You can either have the close paren, or a
        # name.
        while self._parenthesis_stack:
            match = self._expect_match('macro arguments',
                                       MACRO_ARGS_END_PATTERN,
                                       MACRO_ARG_PATTERN)
            self.advance(match.end())
            matchgroups = match.groupdict()
            if matchgroups.get('macro_end') is not None:
                self._parenthesis_stack.pop()
            # we got an argument. let's see what it has
            elif matchgroups.get('value') is not None:
                # we have to process a single macro argument. This mutates
                # the parenthesis stack! If it finds a comma, it will continue
                # the loop.
                self._process_macro_default_arg()
            elif matchgroups.get('more_args') is not None:
                continue
            else:
                raise dbt.exceptions.InternalException(
                    'unhandled regex in _process_macro_args(), no match: {}'
                    .format(matchgroups)
                )
            # if there are more arguments or a macro arg end we'll catch them
            # on the next loop around

    def lex_for_blocks(self):
        while self.data[self.pos:]:
            found = self.find_block()
            if not found:
                break

        raw_toplevel = self.data[self.pos:]
        if len(raw_toplevel) > 0:
            self.blocks.append(BlockData(raw_toplevel))

        return self.blocks




from .analysis import AnalysisParser
from .snapshots import SnapshotParser
from .data_test import DataTestParser
from .docs import DocumentationParser
from .hooks import HookParser
from .macros import MacroParser
from .models import ModelParser
from .schemas import SchemaParser
from .seeds import SeedParser

from .util import ParserUtils

__all__ = [
    'AnalysisParser',
    'SnapshotParser',
    'DataTestParser',
    'DocumentationParser',
    'HookParser',
    'MacroParser',
    'ModelParser',
    'SchemaParser',
    'SeedParser',

    'ParserUtils',
]




import json
from dbt.compat import to_string


class ModelHookType:
    PreHook = 'pre-hook'
    PostHook = 'post-hook'
    Both = [PreHook, PostHook]


def _parse_hook_to_dict(hook_string):
    try:
        hook_dict = json.loads(hook_string)
    except ValueError:
        hook_dict = {"sql": hook_string}

    if 'transaction' not in hook_dict:
        hook_dict['transaction'] = True

    return hook_dict


def get_hook_dict(hook, index):
    if isinstance(hook, dict):
        hook_dict = hook
    else:
        hook_dict = _parse_hook_to_dict(to_string(hook))

    hook_dict['index'] = index
    return hook_dict


def get_hooks(model, hook_key):
    hooks = model.config.get(hook_key, [])

    if not isinstance(hooks, (list, tuple)):
        hooks = [hooks]

    wrapped = [get_hook_dict(hook, i) for i, hook in enumerate(hooks)]
    return wrapped



import abc
from contextlib import contextmanager

import agate
import pytz
import six

import dbt.exceptions
import dbt.flags
import dbt.clients.agate_helper

from dbt.compat import abstractclassmethod, classmethod
from dbt.node_types import NodeType
from dbt.loader import GraphLoader
from dbt.logger import GLOBAL_LOGGER as logger
from dbt.utils import filter_null_values


from dbt.adapters.base.meta import AdapterMeta, available
from dbt.adapters.base import BaseRelation
from dbt.adapters.base import Column
from dbt.adapters.cache import RelationsCache


GET_CATALOG_MACRO_NAME = 'get_catalog'
FRESHNESS_MACRO_NAME = 'collect_freshness'


def _expect_row_value(key, row):
    if key not in row.keys():
        raise dbt.exceptions.InternalException(
            'Got a row without "{}" column, columns: {}'
            .format(key, row.keys())
        )
    return row[key]


def _relations_filter_schemas(schemas):
    def test(row):
        referenced_schema = _expect_row_value('referenced_schema', row)
        dependent_schema = _expect_row_value('dependent_schema', row)
        # handle the null schema
        if referenced_schema is not None:
            referenced_schema = referenced_schema.lower()
        if dependent_schema is not None:
            dependent_schema = dependent_schema.lower()
        return referenced_schema in schemas or dependent_schema in schemas
    return test


def _catalog_filter_schemas(manifest):
    """Return a function that takes a row and decides if the row should be
    included in the catalog output.
    """
    schemas = frozenset((d.lower(), s.lower())
                        for d, s in manifest.get_used_schemas())

    def test(row):
        table_database = _expect_row_value('table_database', row)
        table_schema = _expect_row_value('table_schema', row)
        # the schema may be present but None, which is not an error and should
        # be filtered out
        if table_schema is None:
            return False
        return (table_database.lower(), table_schema.lower()) in schemas
    return test


def _utc(dt, source, field_name):
    """If dt has a timezone, return a new datetime that's in UTC. Otherwise,
    assume the datetime is already for UTC and add the timezone.
    """
    if dt is None:
        raise dbt.exceptions.raise_database_error(
            "Expected a non-null value when querying field '{}' of table "
            " {} but received value 'null' instead".format(
                field_name,
                source))

    elif not hasattr(dt, 'tzinfo'):
        raise dbt.exceptions.raise_database_error(
            "Expected a timestamp value when querying field '{}' of table "
            "{} but received value of type '{}' instead".format(
                field_name,
                source,
                type(dt).__name__))

    elif dt.tzinfo:
        return dt.astimezone(pytz.UTC)
    else:
        return dt.replace(tzinfo=pytz.UTC)


class SchemaSearchMap(dict):
    """A utility class to keep track of what information_schema tables to
    search for what schemas
    """
    def add(self, relation):
        key = relation.information_schema_only()
        if key not in self:
            self[key] = set()
        self[key].add(relation.schema.lower())

    def search(self):
        for information_schema_name, schemas in self.items():
            for schema in schemas:
                yield information_schema_name, schema

    def schemas_searched(self):
        result = set()
        for information_schema_name, schemas in self.items():
            result.update(
                (information_schema_name.database, schema)
                for schema in schemas
            )
        return result

    def flatten(self):
        new = self.__class__()

        database = None
        # iterate once to look for a database name
        seen = {r.database.lower() for r in self if r.database}
        if len(seen) > 1:
            dbt.exceptions.raise_compiler_error(str(seen))
        elif len(seen) == 1:
            database = list(seen)[0]

        for information_schema_name, schema in self.search():
            new.add(information_schema_name.incorporate(
                path={'database': database, 'schema': schema},
                quote_policy={'database': False},
                include_policy={'database': False},
            ))

        return new


@six.add_metaclass(AdapterMeta)
class BaseAdapter(object):
    """The BaseAdapter provides an abstract base class for adapters.

    Adapters must implement the following methods and macros. Some of the
    methods can be safely overridden as a noop, where it makes sense
    (transactions on databases that don't support them, for instance). Those
    methods are marked with a (passable) in their docstrings. Check docstrings
    for type information, etc.

    To implement a macro, implement "${adapter_type}__${macro_name}". in the
    adapter's internal project.

    Methods:
        - exception_handler
        - date_function
        - list_schemas
        - drop_relation
        - truncate_relation
        - rename_relation
        - get_columns_in_relation
        - expand_column_types
        - list_relations_without_caching
        - is_cancelable
        - create_schema
        - drop_schema
        - quote
        - convert_text_type
        - convert_number_type
        - convert_boolean_type
        - convert_datetime_type
        - convert_date_type
        - convert_time_type

    Macros:
        - get_catalog
    """
    requires = {}

    Relation = BaseRelation
    Column = Column
    # This should be an implementation of BaseConnectionManager
    ConnectionManager = None

    # A set of clobber config fields accepted by this adapter
    # for use in materializations
    AdapterSpecificConfigs = frozenset()

    def __init__(self, config):
        self.config = config
        self.cache = RelationsCache()
        self.connections = self.ConnectionManager(config)
        self._internal_manifest_lazy = None

    ###
    # Methods that pass through to the connection manager
    ###
    def acquire_connection(self, name=None):
        return self.connections.set_connection_name(name)

    def release_connection(self):
        return self.connections.release()

    def cleanup_connections(self):
        return self.connections.cleanup_all()

    def clear_transaction(self):
        self.connections.clear_transaction()

    def commit_if_has_connection(self):
        return self.connections.commit_if_has_connection()

    def nice_connection_name(self):
        conn = self.connections.get_thread_connection()
        if conn is None or conn.name is None:
            return '<None>'
        return conn.name

    @contextmanager
    def connection_named(self, name):
        try:
            yield self.acquire_connection(name)
        finally:
            self.release_connection()

    @available.parse(lambda *a, **k: ('', dbt.clients.agate_helper()))
    def execute(self, sql, auto_begin=False, fetch=False):
        """Execute the given SQL. This is a thin wrapper around
        ConnectionManager.execute.

        :param str sql: The sql to execute.
        :param bool auto_begin: If set, and dbt is not currently inside a
            transaction, automatically begin one.
        :param bool fetch: If set, fetch results.
        :return: A tuple of the status and the results (empty if fetch=False).
        :rtype: Tuple[str, agate.Table]
        """
        return self.connections.execute(
            sql=sql,
            auto_begin=auto_begin,
            fetch=fetch
        )

    ###
    # Methods that should never be overridden
    ###
    @classmethod
    def type(cls):
        """Get the type of this adapter. Types must be class-unique and
        consistent.

        :return: The type name
        :rtype: str
        """
        return cls.ConnectionManager.TYPE

    @property
    def _internal_manifest(self):
        if self._internal_manifest_lazy is None:
            manifest = GraphLoader.load_internal(self.config)
            self._internal_manifest_lazy = manifest
        return self._internal_manifest_lazy

    def check_internal_manifest(self):
        """Return the internal manifest (used for executing macros) if it's
        been initialized, otherwise return None.
        """
        return self._internal_manifest_lazy

    ###
    # Caching methods
    ###
    def _schema_is_cached(self, database, schema):
        """Check if the schema is cached, and by default logs if it is not."""

        if dbt.flags.USE_CACHE is False:
            return False
        elif (database, schema) not in self.cache:
            logger.debug(
                'On "{}": cache miss for schema "{}.{}", this is inefficient'
                .format(self.nice_connection_name(), database, schema)
            )
            return False
        else:
            return True

    @classmethod
    def _relations_filter_table(cls, table, schemas):
        """Filter the table as appropriate for relations table entries.
        Subclasses can override this to change filtering rules on a per-adapter
        basis.
        """
        return table.where(_relations_filter_schemas(schemas))

    def _get_cache_schemas(self, manifest, exec_only=False):
        """Get a mapping of each node's "information_schema" relations to a
        set of all schemas expected in that information_schema.

        There may be keys that are technically duplicates on the database side,
        for example all of '"foo", 'foo', '"FOO"' and 'FOO' could coexist as
        databases, and values could overlap as appropriate. All values are
        lowercase strings.
        """
        info_schema_name_map = SchemaSearchMap()
        for node in manifest.nodes.values():
            if exec_only and node.resource_type not in NodeType.executable():
                continue
            relation = self.Relation.create_from(self.config, node)
            info_schema_name_map.add(relation)
        # result is a map whose keys are information_schema Relations without
        # identifiers that have appropriate database prefixes, and whose values
        # are sets of lowercase schema names that are valid members of those
        # schemas
        return info_schema_name_map

    def _relations_cache_for_schemas(self, manifest):
        """Populate the relations cache for the given schemas. Returns an
        iteratble of the schemas populated, as strings.
        """
        if not dbt.flags.USE_CACHE:
            return

        info_schema_name_map = self._get_cache_schemas(manifest,
                                                       exec_only=True)
        for db, schema in info_schema_name_map.search():
            for relation in self.list_relations_without_caching(db, schema):
                self.cache.add(relation)

        # it's possible that there were no relations in some schemas. We want
        # to insert the schemas we query into the cache's `.schemas` attribute
        # so we can check it later
        self.cache.update_schemas(info_schema_name_map.schemas_searched())

    def set_relations_cache(self, manifest, clear=False):
        """Run a query that gets a populated cache of the relations in the
        database and set the cache on this adapter.
        """
        if not dbt.flags.USE_CACHE:
            return

        with self.cache.lock:
            if clear:
                self.cache.clear()
            self._relations_cache_for_schemas(manifest)

    def cache_new_relation(self, relation):
        """Cache a new relation in dbt. It will show up in `list relations`."""
        if relation is None:
            name = self.nice_connection_name()
            dbt.exceptions.raise_compiler_error(
                'Attempted to cache a null relation for {}'.format(name)
            )
        if dbt.flags.USE_CACHE:
            self.cache.add(relation)
        # so jinja doesn't render things
        return ''

    ###
    # Abstract methods for database-specific values, attributes, and types
    ###
    @abstractclassmethod
    def date_function(cls):
        """Get the date function used by this adapter's database.

        :return: The date function
        :rtype: str
        """
        raise dbt.exceptions.NotImplementedException(
            '`date_function` is not implemented for this adapter!')

    @abstractclassmethod
    def is_cancelable(cls):
        raise dbt.exceptions.NotImplementedException(
            '`is_cancelable` is not implemented for this adapter!'
        )

    ###
    # Abstract methods about schemas
    ###
    @abc.abstractmethod
    def list_schemas(self, database):
        """Get a list of existing schemas.

        :param str database: The name of the database to list under.
        :return: All schemas that currently exist in the database
        :rtype: List[str]
        """
        raise dbt.exceptions.NotImplementedException(
            '`list_schemas` is not implemented for this adapter!'
        )

    @available.parse(lambda *a, **k: False)
    def check_schema_exists(self, database, schema):
        """Check if a schema exists.

        The default implementation of this is potentially unnecessarily slow,
        and adapters should implement it if there is an optimized path (and
        there probably is)
        """
        search = (
            s.lower() for s in
            self.list_schemas(database=database)
        )
        return schema.lower() in search

    ###
    # Abstract methods about relations
    ###
    @abc.abstractmethod
    @available.parse_none
    def drop_relation(self, relation):
        """Drop the given relation.

        *Implementors must call self.cache.drop() to preserve cache state!*

        :param self.Relation relation: The relation to drop
        """
        raise dbt.exceptions.NotImplementedException(
            '`drop_relation` is not implemented for this adapter!'
        )

    @abc.abstractmethod
    @available.parse_none
    def truncate_relation(self, relation):
        """Truncate the given relation.

        :param self.Relation relation: The relation to truncate
        """
        raise dbt.exceptions.NotImplementedException(
            '`truncate_relation` is not implemented for this adapter!'
        )

    @abc.abstractmethod
    @available.parse_none
    def rename_relation(self, from_relation, to_relation):
        """Rename the relation from from_relation to to_relation.

        Implementors must call self.cache.rename() to preserve cache state.

        :param self.Relation from_relation: The original relation name
        :param self.Relation to_relation: The new relation name
        """
        raise dbt.exceptions.NotImplementedException(
            '`rename_relation` is not implemented for this adapter!'
        )

    @abc.abstractmethod
    @available.parse_list
    def get_columns_in_relation(self, relation):
        """Get a list of the columns in the given Relation.

        :param self.Relation relation: The relation to query for.
        :return: Information about all columns in the given relation.
        :rtype: List[self.Column]
        """
        raise dbt.exceptions.NotImplementedException(
            '`get_columns_in_relation` is not implemented for this adapter!'
        )

    @available.deprecated('get_columns_in_relation', lambda *a, **k: [])
    def get_columns_in_table(self, schema, identifier):
        """DEPRECATED: Get a list of the columns in the given table."""
        relation = self.Relation.create(
            database=self.config.credentials.database,
            schema=schema,
            identifier=identifier,
            quote_policy=self.config.quoting
        )
        return self.get_columns_in_relation(relation)

    @abc.abstractmethod
    def expand_column_types(self, goal, current):
        """Expand the current table's types to match the goal table. (passable)

        :param self.Relation goal: A relation that currently exists in the
            database with columns of the desired types.
        :param self.Relation current: A relation that currently exists in the
            database with columns of unspecified types.
        """
        raise dbt.exceptions.NotImplementedException(
            '`expand_target_column_types` is not implemented for this adapter!'
        )

    @abc.abstractmethod
    def list_relations_without_caching(self, information_schema, schema):
        """List relations in the given schema, bypassing the cache.

        This is used as the underlying behavior to fill the cache.

        :param Relation information_schema: The information schema to list
            relations from.
        :param str schema: The name of the schema to list relations from.
        :return: The relations in schema
        :rtype: List[self.Relation]
        """
        raise dbt.exceptions.NotImplementedException(
            '`list_relations_without_caching` is not implemented for this '
            'adapter!'
        )

    ###
    # Provided methods about relations
    ###
    @available.parse_list
    def get_missing_columns(self, from_relation, to_relation):
        """Returns a list of Columns in from_relation that are missing from
        to_relation.

        :param Relation from_relation: The relation that might have extra
            columns
        :param Relation to_relation: The realtion that might have columns
            missing
        :return: The columns in from_relation that are missing from to_relation
        :rtype: List[self.Relation]
        """
        if not isinstance(from_relation, self.Relation):
            dbt.exceptions.invalid_type_error(
                method_name='get_missing_columns',
                arg_name='from_relation',
                got_value=from_relation,
                expected_type=self.Relation)

        if not isinstance(to_relation, self.Relation):
            dbt.exceptions.invalid_type_error(
                method_name='get_missing_columns',
                arg_name='to_relation',
                got_value=to_relation,
                expected_type=self.Relation)

        from_columns = {
            col.name: col for col in
            self.get_columns_in_relation(from_relation)
        }

        to_columns = {
            col.name: col for col in
            self.get_columns_in_relation(to_relation)
        }

        missing_columns = set(from_columns.keys()) - set(to_columns.keys())

        return [
            col for (col_name, col) in from_columns.items()
            if col_name in missing_columns
        ]

    @available.parse_none
    def valid_snapshot_target(self, relation):
        """Ensure that the target relation is valid, by making sure it has the
        expected columns.

        :param Relation relation: The relation to check
        :raises dbt.exceptions.CompilationException: If the columns are
            incorrect.
        """
        if not isinstance(relation, self.Relation):
            dbt.exceptions.invalid_type_error(
                method_name='valid_snapshot_target',
                arg_name='relation',
                got_value=relation,
                expected_type=self.Relation)

        columns = self.get_columns_in_relation(relation)
        names = set(c.name.lower() for c in columns)
        expanded_keys = ('scd_id', 'valid_from', 'valid_to')
        extra = []
        missing = []
        for legacy in expanded_keys:
            desired = 'dbt_' + legacy
            if desired not in names:
                missing.append(desired)
                if legacy in names:
                    extra.append(legacy)

        if missing:
            if extra:
                msg = (
                    'Snapshot target has ("{}") but not ("{}") - is it an '
                    'unmigrated previous version archive?'
                    .format('", "'.join(extra), '", "'.join(missing))
                )
            else:
                msg = (
                    'Snapshot target is not a snapshot table (missing "{}")'
                    .format('", "'.join(missing))
                )
            dbt.exceptions.raise_compiler_error(msg)

    @available.parse_none
    def expand_target_column_types(self, from_relation, to_relation):
        if not isinstance(from_relation, self.Relation):
            dbt.exceptions.invalid_type_error(
                method_name='expand_target_column_types',
                arg_name='from_relation',
                got_value=from_relation,
                expected_type=self.Relation)

        if not isinstance(to_relation, self.Relation):
            dbt.exceptions.invalid_type_error(
                method_name='expand_target_column_types',
                arg_name='to_relation',
                got_value=to_relation,
                expected_type=self.Relation)

        self.expand_column_types(from_relation, to_relation)

    def list_relations(self, database, schema):
        if self._schema_is_cached(database, schema):
            return self.cache.get_relations(database, schema)

        information_schema = self.Relation.create(
            database=database,
            schema=schema,
            model_name='',
            quote_policy=self.config.quoting
        ).information_schema()

        # we can't build the relations cache because we don't have a
        # manifest so we can't run any operations.
        relations = self.list_relations_without_caching(
            information_schema, schema
        )

        logger.debug('with database={}, schema={}, relations={}'
                     .format(database, schema, relations))
        return relations

    def _make_match_kwargs(self, database, schema, identifier):
        quoting = self.config.quoting
        if identifier is not None and quoting['identifier'] is False:
            identifier = identifier.lower()

        if schema is not None and quoting['schema'] is False:
            schema = schema.lower()

        if database is not None and quoting['database'] is False:
            database = database.lower()

        return filter_null_values({
            'database': database,
            'identifier': identifier,
            'schema': schema,
        })

    def _make_match(self, relations_list, database, schema, identifier):

        matches = []

        search = self._make_match_kwargs(database, schema, identifier)

        for relation in relations_list:
            if relation.matches(**search):
                matches.append(relation)

        return matches

    @available.parse_none
    def get_relation(self, database, schema, identifier):
        relations_list = self.list_relations(database, schema)

        matches = self._make_match(relations_list, database, schema,
                                   identifier)

        if len(matches) > 1:
            kwargs = {
                'identifier': identifier,
                'schema': schema,
                'database': database,
            }
            dbt.exceptions.get_relation_returned_multiple_results(
                kwargs, matches
            )

        elif matches:
            return matches[0]

        return None

    @available.deprecated('get_relation', lambda *a, **k: False)
    def already_exists(self, schema, name):
        """DEPRECATED: Return if a model already exists in the database"""
        database = self.config.credentials.database
        relation = self.get_relation(database, schema, name)
        return relation is not None

    ###
    # ODBC FUNCTIONS -- these should not need to change for every adapter,
    #                   although some adapters may override them
    ###
    @abc.abstractmethod
    @available.parse_none
    def create_schema(self, database, schema):
        """Create the given schema if it does not exist.

        :param str schema: The schema name to create.
        """
        raise dbt.exceptions.NotImplementedException(
            '`create_schema` is not implemented for this adapter!'
        )

    @abc.abstractmethod
    def drop_schema(self, database, schema):
        """Drop the given schema (and everything in it) if it exists.

        :param str schema: The schema name to drop.
        """
        raise dbt.exceptions.NotImplementedException(
            '`drop_schema` is not implemented for this adapter!'
        )

    @available
    @abstractclassmethod
    def quote(cls, identifier):
        """Quote the given identifier, as appropriate for the database.

        :param str identifier: The identifier to quote
        :return: The quoted identifier
        :rtype: str
        """
        raise dbt.exceptions.NotImplementedException(
            '`quote` is not implemented for this adapter!'
        )

    @available
    def quote_as_configured(self, identifier, quote_key):
        """Quote or do not quote the given identifer as configured in the
        project config for the quote key.

        The quote key should be one of 'database' (on bigquery, 'profile'),
        'identifier', or 'schema', or it will be treated as if you set `True`.
        """
        default = self.Relation.DEFAULTS['quote_policy'].get(quote_key)
        if self.config.quoting.get(quote_key, default):
            return self.quote(identifier)
        else:
            return identifier

    ###
    # Conversions: These must be implemented by concrete implementations, for
    # converting agate types into their sql equivalents.
    ###
    @abstractclassmethod
    def convert_text_type(cls, agate_table, col_idx):
        """Return the type in the database that best maps to the agate.Text
        type for the given agate table and column index.

        :param agate.Table agate_table: The table
        :param int col_idx: The index into the agate table for the column.
        :return: The name of the type in the database
        :rtype: str
        """
        raise dbt.exceptions.NotImplementedException(
            '`convert_text_type` is not implemented for this adapter!')

    @abstractclassmethod
    def convert_number_type(cls, agate_table, col_idx):
        """Return the type in the database that best maps to the agate.Number
        type for the given agate table and column index.

        :param agate.Table agate_table: The table
        :param int col_idx: The index into the agate table for the column.
        :return: The name of the type in the database
        :rtype: str
        """
        raise dbt.exceptions.NotImplementedException(
            '`convert_number_type` is not implemented for this adapter!')

    @abstractclassmethod
    def convert_boolean_type(cls, agate_table, col_idx):
        """Return the type in the database that best maps to the agate.Boolean
        type for the given agate table and column index.

        :param agate.Table agate_table: The table
        :param int col_idx: The index into the agate table for the column.
        :return: The name of the type in the database
        :rtype: str
        """
        raise dbt.exceptions.NotImplementedException(
            '`convert_boolean_type` is not implemented for this adapter!')

    @abstractclassmethod
    def convert_datetime_type(cls, agate_table, col_idx):
        """Return the type in the database that best maps to the agate.DateTime
        type for the given agate table and column index.

        :param agate.Table agate_table: The table
        :param int col_idx: The index into the agate table for the column.
        :return: The name of the type in the database
        :rtype: str
        """
        raise dbt.exceptions.NotImplementedException(
            '`convert_datetime_type` is not implemented for this adapter!')

    @abstractclassmethod
    def convert_date_type(cls, agate_table, col_idx):
        """Return the type in the database that best maps to the agate.Date
        type for the given agate table and column index.

        :param agate.Table agate_table: The table
        :param int col_idx: The index into the agate table for the column.
        :return: The name of the type in the database
        :rtype: str
        """
        raise dbt.exceptions.NotImplementedException(
            '`convert_date_type` is not implemented for this adapter!')

    @abstractclassmethod
    def convert_time_type(cls, agate_table, col_idx):
        """Return the type in the database that best maps to the
        agate.TimeDelta type for the given agate table and column index.

        :param agate.Table agate_table: The table
        :param int col_idx: The index into the agate table for the column.
        :return: The name of the type in the database
        :rtype: str
        """
        raise dbt.exceptions.NotImplementedException(
            '`convert_time_type` is not implemented for this adapter!')

    @available
    @classmethod
    def convert_type(cls, agate_table, col_idx):
        return cls.convert_agate_type(agate_table, col_idx)

    @classmethod
    def convert_agate_type(cls, agate_table, col_idx):
        agate_type = agate_table.column_types[col_idx]
        conversions = [
            (agate.Text, cls.convert_text_type),
            (agate.Number, cls.convert_number_type),
            (agate.Boolean, cls.convert_boolean_type),
            (agate.DateTime, cls.convert_datetime_type),
            (agate.Date, cls.convert_date_type),
            (agate.TimeDelta, cls.convert_time_type),
        ]
        for agate_cls, func in conversions:
            if isinstance(agate_type, agate_cls):
                return func(agate_table, col_idx)

    ###
    # Operations involving the manifest
    ###
    def execute_macro(self, macro_name, manifest=None, project=None,
                      context_override=None, kwargs=None, release=False):
        """Look macro_name up in the manifest and execute its results.

        :param str macro_name: The name of the macro to execute.
        :param Optional[Manifest] manifest: The manifest to use for generating
            the base macro execution context. If none is provided, use the
            internal manifest.
        :param Optional[str] project: The name of the project to search in, or
            None for the first match.
        :param Optional[dict] context_override: An optional dict to update()
            the macro execution context.
        :param Optional[dict] kwargs: An optional dict of keyword args used to
            pass to the macro.
        :param bool release: If True, release the connection after executing.

        Return an an AttrDict with three attributes: 'table', 'data', and
            'status'. 'table' is an agate.Table.
        """
        if kwargs is None:
            kwargs = {}
        if context_override is None:
            context_override = {}

        if manifest is None:
            manifest = self._internal_manifest

        macro = manifest.find_macro_by_name(macro_name, project)
        if macro is None:
            if project is None:
                package_name = 'any package'
            else:
                package_name = 'the "{}" package'.format(project)

            # The import of dbt.context.runtime below shadows 'dbt'
            import dbt.exceptions
            raise dbt.exceptions.RuntimeException(
                'dbt could not find a macro with the name "{}" in {}'
                .format(macro_name, package_name)
            )
        # This causes a reference cycle, as dbt.context.runtime.generate()
        # ends up calling get_adapter, so the import has to be here.
        import dbt.context.operation
        macro_context = dbt.context.operation.generate(
            macro,
            self.config,
            manifest
        )
        macro_context.update(context_override)

        macro_function = macro.generator(macro_context)

        try:
            result = macro_function(**kwargs)
        finally:
            if release:
                self.release_connection()
        return result

    @classmethod
    def _catalog_filter_table(cls, table, manifest):
        """Filter the table as appropriate for catalog entries. Subclasses can
        override this to change filtering rules on a per-adapter basis.
        """
        return table.where(_catalog_filter_schemas(manifest))

    def get_catalog(self, manifest):
        """Get the catalog for this manifest by running the get catalog macro.
        Returns an agate.Table of catalog information.
        """
        information_schemas = list(self._get_cache_schemas(manifest).keys())
        # make it a list so macros can index into it.
        kwargs = {'information_schemas': information_schemas}
        table = self.execute_macro(GET_CATALOG_MACRO_NAME,
                                   kwargs=kwargs,
                                   release=True)

        results = self._catalog_filter_table(table, manifest)
        return results

    def cancel_open_connections(self):
        """Cancel all open connections."""
        return self.connections.cancel_open()

    def calculate_freshness(self, source, loaded_at_field, manifest=None):
        """Calculate the freshness of sources in dbt, and return it"""
        # in the future `source` will be a Relation instead of a string
        kwargs = {
            'source': source,
            'loaded_at_field': loaded_at_field
        }

        # run the macro
        table = self.execute_macro(
            FRESHNESS_MACRO_NAME,
            kwargs=kwargs,
            release=True,
            manifest=manifest
        )
        # now we have a 1-row table of the maximum `loaded_at_field` value and
        # the current time according to the db.
        if len(table) != 1 or len(table[0]) != 2:
            dbt.exceptions.raise_compiler_error(
                'Got an invalid result from "{}" macro: {}'.format(
                    FRESHNESS_MACRO_NAME, [tuple(r) for r in table]
                )
            )

        max_loaded_at = _utc(table[0][0], source, loaded_at_field)
        snapshotted_at = _utc(table[0][1], source, loaded_at_field)

        age = (snapshotted_at - max_loaded_at).total_seconds()
        return {
            'max_loaded_at': max_loaded_at,
            'snapshotted_at': snapshotted_at,
            'age': age,
        }



from dbt.api.object import APIObject
from dbt.contracts.common import named_property


CONNECTION_CONTRACT = {
    'type': 'object',
    'additionalProperties': False,
    'properties': {
        'type': {
            'type': 'string',
            # valid python identifiers only
            'pattern': r'^[A-Za-z_][A-Za-z0-9_]+$',
        },
        'name': {
            'type': ['null', 'string'],
        },
        'state': {
            'enum': ['init', 'open', 'closed', 'fail'],
        },
        'transaction_open': {
            'type': 'boolean',
        },
        # we can't serialize this so we can't require it as part of the
        # contract.
        # 'handle': {
        #     'type': ['null', 'object'],
        # },
        # credentials are validated separately by the adapter packages
        'credentials': {
            'description': (
                'The credentials object here should match the connection type.'
            ),
            'type': 'object',
            'additionalProperties': True,
        }
    },
    'required': [
        'type', 'name', 'state', 'transaction_open', 'credentials'
    ],
}


class Connection(APIObject):
    SCHEMA = CONNECTION_CONTRACT

    def __init__(self, credentials, *args, **kwargs):
        # we can't serialize handles
        self._handle = kwargs.pop('handle')
        super(Connection, self).__init__(credentials=credentials.serialize(),
                                         *args, **kwargs)
        # this will validate itself in its own __init__.
        self._credentials = credentials

    @property
    def credentials(self):
        return self._credentials

    @property
    def handle(self):
        return self._handle

    @handle.setter
    def handle(self, value):
        self._handle = value

    name = named_property('name', 'The name of this connection')
    state = named_property('state', 'The state of the connection')
    transaction_open = named_property(
        'transaction_open',
        'True if there is an open transaction, False otherwise.'
    )



STRICT_MODE = False
FULL_REFRESH = False
USE_CACHE = True
WARN_ERROR = False
TEST_NEW_PARSER = False


def reset():
    global STRICT_MODE, FULL_REFRESH, USE_CACHE, WARN_ERROR, TEST_NEW_PARSER

    STRICT_MODE = False
    FULL_REFRESH = False
    USE_CACHE = True
    WARN_ERROR = False
    TEST_NEW_PARSER = False


def set_from_args(args):
    global STRICT_MODE, FULL_REFRESH, USE_CACHE, WARN_ERROR, TEST_NEW_PARSER
    USE_CACHE = getattr(args, 'use_cache', True)

    FULL_REFRESH = getattr(args, 'full_refresh', False)
    STRICT_MODE = getattr(args, 'strict', False)
    WARN_ERROR = (
        STRICT_MODE or
        getattr(args, 'warn_error', False)
    )

    TEST_NEW_PARSER = getattr(args, 'test_new_parser', False)



from dbt.logger import GLOBAL_LOGGER as logger
from dbt.task.base import ConfiguredTask
from dbt.adapters.factory import get_adapter
from dbt.loader import GraphLoader

import dbt
import dbt.utils
import dbt.exceptions


class RunOperationTask(ConfiguredTask):
    def _get_macro_parts(self):
        macro_name = self.args.macro
        if '.' in macro_name:
            package_name, macro_name = macro_name.split(".", 1)
        else:
            package_name = None

        return package_name, macro_name

    def _get_kwargs(self):
        return dbt.utils.parse_cli_vars(self.args.args)

    def _run_unsafe(self):
        manifest = GraphLoader.load_all(self.config)
        adapter = get_adapter(self.config)

        package_name, macro_name = self._get_macro_parts()
        macro_kwargs = self._get_kwargs()

        with adapter.connection_named('macro_{}'.format(macro_name)):
            adapter.clear_transaction()
            res = adapter.execute_macro(
                macro_name,
                project=package_name,
                kwargs=macro_kwargs,
                manifest=manifest
            )

        return res

    def run(self):
        try:
            result = self._run_unsafe()
        except dbt.exceptions.Exception as exc:
            logger.error(
                'Encountered an error while running operation: {}'
                .format(exc)
            )
            logger.debug('', exc_info=True)
            return False, None
        except Exception as exc:
            logger.error(
                'Encountered an uncaught exception while running operation: {}'
                .format(exc)
            )
            logger.debug('', exc_info=True)
            return False, None
        else:
            return True, result

    def interpret_results(self, results):
        success, _ = results
        return success



import itertools
import os

from dbt.task.base import BaseTask

from dbt.config import RuntimeConfig
from dbt.adapters.factory import get_adapter
from dbt.logger import GLOBAL_LOGGER as logger
from dbt.clients import system
from dbt.exceptions import RuntimeException


_UNQUOTE_RENAME_COLUMNS = (
    ('valid_from', 'dbt_valid_from'),
    ('valid_to', 'dbt_valid_to'),
    ('scd_id', 'dbt_scd_id'),
)


SNAPSHOT_TEMPLATE = '''
{{% snapshot {name} %}}
    {{{{
        config({kwargs})
    }}}}
    select * from {source_relation}
{{% endsnapshot %}}
'''


class Migrator(object):
    """Migrate a single archive config"""
    def __init__(self, manager, adapter, archive_def):
        self.manager = manager
        self.archive_def = archive_def
        self.adapter = adapter
        self.relation = adapter.Relation.create(
            database=self.archive_def['target_database'],
            schema=self.archive_def['target_schema'],
            identifier=self.archive_def['target_table'],
            quote_policy=self.manager.config.quoting
        )
        self.backup = self.append_relation_name()
        self.snapshot_path = os.path.join(self.manager.snapshot_root,
                                          self.relation.identifier + '.sql')

    def get_renamed_columns(self, quote=True):
        columns = []

        unquote = _UNQUOTE_RENAME_COLUMNS
        if self.adapter.type() == 'snowflake':
            unquote = itertools.chain(
                _UNQUOTE_RENAME_COLUMNS,
                [('dbt_updated_at', 'dbt_updated_at')]
            )

        for old, new in unquote:
            if quote:
                old = self.adapter.quote(old)
            columns.append((old, new))
        return columns

    def append_relation_name(self, end='_dbt_archive_migration_backup'):
        new_name = self.relation.identifier + end
        backup = self.relation.incorporate(
            path={'identifier': new_name}, table_name=new_name
        )
        return backup

    def migrate_archive_ctas(self, dest):
        # get the columns
        columns = self.adapter.get_columns_in_relation(self.relation)
        if len(columns) == 0:
            # the archive target must not exist? Continue, that is ok.
            logger.info('  - Table {} does not exist, nothing to migrate.'
                        .format(self.relation))
            return
        cols = {c.name.lower(): c.name for c in columns}
        renames = self.get_renamed_columns()
        select_parts = []
        select_as_parts = []
        for old, new in renames:
            key = old.strip('"').lower()
            if key not in cols:
                raise Exception(
                    'expected column like {} not but it is not in the table!'
                    .format(key)
                )
            del cols[key]
            select_as_parts.append('{} as {}'.format(old, new))

        for column in columns:
            name = column.name
            if name.lower() in cols:
                select_parts.append(self.adapter.quote(name))

        selections = ', '.join(itertools.chain(select_parts, select_as_parts))
        ctas = 'create table {!s} as (select {} from {!s})'.format(
            dest, selections, self.relation
        )
        self.adapter.execute(ctas)

    def migrate_archive_postgres(self):
        """Migrate the archive using "alter table" commands to rename columns.
        """
        self.adapter.connections.add_begin_query()
        tmp = self.append_relation_name('_dbt_archive_migration_tmp')
        logger.debug('  - Making new archive at {}'.format(tmp))
        self.migrate_archive_ctas(dest=tmp)
        logger.info('  - Backing up table to {}'.format(self.backup))
        self.adapter.rename_relation(self.relation, self.backup)
        logger.debug('  - Renaming temp archive to final')
        self.adapter.rename_relation(tmp, self.relation)
        self.adapter.connections.add_commit_query()

    def migrate_archive_snowflake(self):
        """Migrate the archive by create table as select ..."""
        logger.debug('  - Making new archive at {}'.format(self.backup))
        self.migrate_archive_ctas(dest=self.backup)
        logger.info('  - Backing up table to {}'.format(self.backup))
        self.adapter.execute('alter table {!s} swap with {!s}'.format(
            self.relation, self.backup
        ))

    def migrate_archive_bigquery(self):
        """Migrate the archive by select * EXCEPT(...) into itself
        """
        logger.info('  - Backing up table to {}'.format(self.backup))
        self.adapter.execute(
            'create table {} as (select * from {})'
            .format(self.backup, self.relation)
        )
        columns = self.get_renamed_columns()
        except_str = ', '.join(o for o, _ in columns)
        rename_str = ', '.join('{} as {}'.format(o, n) for o, n in columns)
        sql = 'select * EXCEPT({}), {} from {!s}'.format(
            except_str, rename_str, self.relation
        )
        self.adapter.connections.create_table(
            database=self.relation.database,
            schema=self.relation.schema,
            table_name=self.relation.identifier,
            sql=sql
        )

    def migrate_archive_table(self):
        logger.info('  - Starting table migration')

        if self.adapter.type() == 'bigquery':
            self.migrate_archive_bigquery()
        elif self.adapter.type() == 'snowflake':
            self.migrate_archive_snowflake()
        else:
            self.migrate_archive_postgres()
        logger.info('  - Finished table migration')

    def build_archive_data(self):
        source_relation = [
            self.adapter.quote_as_configured(
                self.archive_def['source_database'], 'database'
            ),
            self.adapter.quote_as_configured(
                self.archive_def['source_schema'],
                'schema'
            ),
            self.adapter.quote_as_configured(
                self.archive_def['source_table'],
                'identifier'
            ),
        ]

        kwargs = {
            'target_database': self.archive_def['target_database'],
            'target_schema': self.archive_def['target_schema'],
            'updated_at': self.archive_def['updated_at'],
            'strategy': 'timestamp',
            'unique_key': self.archive_def['unique_key'],
        }

        return SNAPSHOT_TEMPLATE.format(
            source_relation='.'.join(source_relation),
            kwargs=repr(kwargs),
            name=self.archive_def['target_table']
        )

    def write_file(self):
        logger.debug('  - Writing snapshot to {}'.format(self.snapshot_path))
        contents = self.build_archive_data()

        wrote = system.make_file(path=self.snapshot_path, contents=contents)
        if wrote:
            logger.info('  - Wrote new snapshot file to {}'
                        .format(self.snapshot_path))
        else:
            logger.error('  - Error: Could not write new snapshot file to {}'
                         .format(self.snapshot_path))


class ArchiveOkConfig(RuntimeConfig):
    @classmethod
    def from_args(cls, args, allow_archive_configs=True):
        supercls = super(ArchiveOkConfig, cls)
        return supercls.from_args(
            args=args,
            allow_archive_configs=allow_archive_configs
        )


class MigrationTask(BaseTask):
    ConfigType = ArchiveOkConfig

    def __init__(self, args, config):
        if not args.from_archive:
            raise RuntimeException(
                'The --from-archive paramteter is required!'
            )
        if args.apply:
            args.write_files = True
            args.migrate_database = True
        super(MigrationTask, self).__init__(args, config)

        self.snapshot_root = os.path.normpath(self.config.snapshot_paths[0])
        system.make_directory(self.snapshot_root)

        self.backups_made = []
        self.snapshots_written = []

    def archive_definitions(self, adapter):
        default_database = self.config.credentials.database
        for archive in self.config.archive:
            target_database = archive.get('target_database', default_database)
            target_schema = archive['target_schema']
            source_database = archive.get('source_database', default_database)
            source_schema = archive['source_schema']
            for table in archive['tables']:
                table_copy = table.copy()
                table_copy['target_database'] = target_database
                table_copy['target_schema'] = target_schema
                table_copy['source_database'] = source_database
                table_copy['source_schema'] = source_schema

                yield Migrator(self, adapter, table_copy)

    def perform_migration_with(self, adapter):
        archive_defs = list(self.archive_definitions(adapter))
        num_defs = len(archive_defs)
        logger.info('Found {} archive{} to migrate'.format(
            num_defs, '' if num_defs == 1 else 's'
        ))
        logger.info('')
        for idx, migrator in enumerate(archive_defs, start=1):
            logger.info('Archive {} of {}: {}'
                        .format(idx, len(archive_defs), migrator.relation))
            if self.args.migrate_database:
                migrator.migrate_archive_table()
                self.backups_made.append(migrator.backup)
            else:
                logger.info('  - Skipping migration in dry-run mode')
            if self.args.write_files:
                migrator.write_file()
                self.snapshots_written.append(migrator.snapshot_path)
            else:
                logger.info('  - Skipping new snapshot file in dry-run mode')
            logger.info('')

    def run(self):
        adapter = get_adapter(self.config)
        with adapter.connection_named('migration'):
            self.perform_migration_with(adapter)

        if self.backups_made:
            logger.info('The following backup tables were created:')
            for backup in self.backups_made:
                logger.info('  - {!s}'.format(backup))
            logger.info('')

        if self.snapshots_written:
            logger.info('The following snapshot files were created:')
            for written in self.snapshots_written:
                logger.info('  - {!s}'.format(written))
            logger.info('')

        applied_something = any((self.args.apply,
                                 self.args.migrate_database,
                                 self.args.write_files))
        if applied_something:
            msg = (
                'After verifying the migrated tables in the database, please '
                'drop the backup\ntables and remove any archive configs from '
                'your dbt_project.yml file.'
            )
        else:
            msg = 'Re-run this script with `--apply` to apply these migrations'
        logger.info(msg)



import codecs
import linecache
import os

import jinja2
import jinja2._compat
import jinja2.ext
import jinja2.nodes
import jinja2.parser
import jinja2.sandbox

import dbt.compat
import dbt.exceptions
import dbt.utils

from dbt.clients._jinja_blocks import BlockIterator

from dbt.logger import GLOBAL_LOGGER as logger  # noqa


class MacroFuzzParser(jinja2.parser.Parser):
    def parse_macro(self):
        node = jinja2.nodes.Macro(lineno=next(self.stream).lineno)

        # modified to fuzz macros defined in the same file. this way
        # dbt can understand the stack of macros being called.
        #  - @cmcarthur
        node.name = dbt.utils.get_dbt_macro_name(
            self.parse_assign_target(name_only=True).name)

        self.parse_signature(node)
        node.body = self.parse_statements(('name:endmacro',),
                                          drop_needle=True)
        return node


class MacroFuzzEnvironment(jinja2.sandbox.SandboxedEnvironment):
    def _parse(self, source, name, filename):
        return MacroFuzzParser(
            self, source, name,
            jinja2._compat.encode_filename(filename)
        ).parse()

    def _compile(self, source, filename):
        """Override jinja's compilation to stash the rendered source inside
        the python linecache for debugging.
        """
        if filename == '<template>':
            # make a better filename
            filename = 'dbt-{}'.format(
                codecs.encode(os.urandom(12), 'hex').decode('ascii')
            )
            # encode, though I don't think this matters
            filename = jinja2._compat.encode_filename(filename)
            # put ourselves in the cache
            linecache.cache[filename] = (
                len(source),
                None,
                [line + '\n' for line in source.splitlines()],
                filename
            )

        return super(MacroFuzzEnvironment, self)._compile(source, filename)


class TemplateCache(object):

    def __init__(self):
        self.file_cache = {}

    def get_node_template(self, node):
        key = (node['package_name'], node['original_file_path'])

        if key in self.file_cache:
            return self.file_cache[key]

        template = get_template(
            string=node.get('raw_sql'),
            ctx={},
            node=node
        )
        self.file_cache[key] = template

        return template

    def clear(self):
        self.file_cache.clear()


template_cache = TemplateCache()


def macro_generator(node):
    def apply_context(context):
        def call(*args, **kwargs):
            name = node.get('name')
            template = template_cache.get_node_template(node)
            module = template.make_module(context, False, context)

            macro = module.__dict__[dbt.utils.get_dbt_macro_name(name)]
            module.__dict__.update(context)

            try:
                return macro(*args, **kwargs)
            except dbt.exceptions.MacroReturn as e:
                return e.value
            except (TypeError, jinja2.exceptions.TemplateRuntimeError) as e:
                dbt.exceptions.raise_compiler_error(str(e), node)
            except dbt.exceptions.CompilationException as e:
                e.stack.append(node)
                raise e

        return call
    return apply_context


class MaterializationExtension(jinja2.ext.Extension):
    tags = ['materialization']

    def parse(self, parser):
        node = jinja2.nodes.Macro(lineno=next(parser.stream).lineno)
        materialization_name = \
            parser.parse_assign_target(name_only=True).name

        adapter_name = 'default'
        node.args = []
        node.defaults = []

        while parser.stream.skip_if('comma'):
            target = parser.parse_assign_target(name_only=True)

            if target.name == 'default':
                pass

            elif target.name == 'adapter':
                parser.stream.expect('assign')
                value = parser.parse_expression()
                adapter_name = value.value

            else:
                dbt.exceptions.invalid_materialization_argument(
                    materialization_name, target.name)

        node.name = dbt.utils.get_materialization_macro_name(
            materialization_name, adapter_name)

        node.body = parser.parse_statements(('name:endmaterialization',),
                                            drop_needle=True)

        return node


class DocumentationExtension(jinja2.ext.Extension):
    tags = ['docs']

    def parse(self, parser):
        node = jinja2.nodes.Macro(lineno=next(parser.stream).lineno)
        docs_name = parser.parse_assign_target(name_only=True).name

        node.args = []
        node.defaults = []
        node.name = dbt.utils.get_docs_macro_name(docs_name)
        node.body = parser.parse_statements(('name:enddocs',),
                                            drop_needle=True)
        return node


def _is_dunder_name(name):
    return name.startswith('__') and name.endswith('__')


def create_macro_capture_env(node):

    class ParserMacroCapture(jinja2.Undefined):
        """
        This class sets up the parser to capture macros.
        """
        def __init__(self, hint=None, obj=None, name=None, exc=None):
            super(ParserMacroCapture, self).__init__(hint=hint, name=name)
            self.node = node
            self.name = name
            self.package_name = node.get('package_name')
            # jinja uses these for safety, so we have to override them.
            # see https://github.com/pallets/jinja/blob/master/jinja2/sandbox.py#L332-L339 # noqa
            self.unsafe_callable = False
            self.alters_data = False

        def __deepcopy__(self, memo):
            path = os.path.join(self.node.get('root_path'),
                                self.node.get('original_file_path'))

            logger.debug(
                'dbt encountered an undefined variable, "{}" in node {}.{} '
                '(source path: {})'
                .format(self.name, self.node.get('package_name'),
                        self.node.get('name'), path))

            # match jinja's message
            dbt.exceptions.raise_compiler_error(
                "{!r} is undefined".format(self.name),
                node=self.node
            )

        def __getitem__(self, name):
            # Propagate the undefined value if a caller accesses this as if it
            # were a dictionary
            return self

        def __getattr__(self, name):
            if name == 'name' or _is_dunder_name(name):
                raise AttributeError(
                    "'{}' object has no attribute '{}'"
                    .format(type(self).__name__, name)
                )

            self.package_name = self.name
            self.name = name

            return self

        def __call__(self, *args, **kwargs):
            return self

    return ParserMacroCapture


def create_warn_undefined_env(node):
    class WarnUndefined(jinja2.Undefined):
        __slots__ = ('node',)

        def __init__(self, hint=None, obj=None, name=None, exc=None):
            super(WarnUndefined, self).__init__(hint=hint, name=name)
            self.node = node

        def _fail_or_warn_with_undefined_error(self, *args, **kwargs):
            try:
                self._fail_with_undefined_error(*args, **kwargs)
            except Exception as exc:
                if dbt.flags.WARN_ERROR:
                    raise
                else:
                    msg = 'Compilation warning in {}:\n\t{}'
                    if self.node is None:
                        model_desc = 'rendering'
                    else:
                        model_desc = "{} {} ({})".format(
                            self.node.get('resource_type'),
                            self.node.get('name', 'unknown'),
                            self.node.get('original_file_path')
                        )
                    logger.warning(msg.format(model_desc, str(exc)))

        def __eq__(self, other):
            self._fail_or_warn_with_undefined_error(other)

            return type(self) is type(other)

        def __ne__(self, other):
            self._fail_or_warn_with_undefined_error(other)

            return not self.__eq__(other)

        def __hash__(self):
            self._fail_or_warn_with_undefined_error()

            return id(type(self))

        def __str__(self):
            self._fail_or_warn_with_undefined_error()

            return u''

        def __len__(self):
            self._fail_or_warn_with_undefined_error()

            return 0

        def __iter__(self):
            self._fail_or_warn_with_undefined_error()

            if 0:
                yield None

        def __nonzero__(self):
            self._fail_or_warn_with_undefined_error()

            return False
        __bool__ = __nonzero__

        def __repr__(self):
            return 'Undefined'

    return WarnUndefined


def get_environment(node=None, capture_macros=False):
    args = {
        'extensions': ['jinja2.ext.do']
    }

    if capture_macros:
        args['undefined'] = create_macro_capture_env(node)
    else:
        args['undefined'] = create_warn_undefined_env(node)

    args['extensions'].append(MaterializationExtension)
    args['extensions'].append(DocumentationExtension)

    return MacroFuzzEnvironment(**args)


def parse(string):
    try:
        return get_environment().parse(dbt.compat.to_string(string))

    except (jinja2.exceptions.TemplateSyntaxError,
            jinja2.exceptions.UndefinedError) as e:
        e.translated = False
        dbt.exceptions.raise_compiler_error(str(e))


def get_template(string, ctx, node=None, capture_macros=False):
    try:
        env = get_environment(node, capture_macros)

        template_source = dbt.compat.to_string(string)
        return env.from_string(template_source, globals=ctx)

    except (jinja2.exceptions.TemplateSyntaxError,
            jinja2.exceptions.UndefinedError) as e:
        e.translated = False
        dbt.exceptions.raise_compiler_error(str(e), node)


def render_template(template, ctx, node=None):
    try:
        return template.render(ctx)

    except (jinja2.exceptions.TemplateSyntaxError,
            jinja2.exceptions.UndefinedError) as e:
        e.translated = False
        dbt.exceptions.raise_compiler_error(str(e), node)


def get_rendered(string, ctx, node=None,
                 capture_macros=False):
    template = get_template(string, ctx, node,
                            capture_macros=capture_macros)

    return render_template(template, ctx, node)


def undefined_error(msg):
    raise jinja2.exceptions.UndefinedError(msg)


def extract_toplevel_blocks(data):
    return BlockIterator(data).lex_for_blocks()



import os
import shutil

from dbt.adapters.factory import get_adapter
from dbt.clients.system import write_json
from dbt.compat import bigint
from dbt.include.global_project import DOCS_INDEX_FILE_PATH
import dbt.ui.printer
import dbt.utils
import dbt.compilation
import dbt.exceptions

from dbt.task.compile import CompileTask


CATALOG_FILENAME = 'catalog.json'


def get_stripped_prefix(source, prefix):
    """Go through source, extracting every key/value pair where the key starts
    with the given prefix.
    """
    cut = len(prefix)
    return {
        k[cut:]: v for k, v in source.items()
        if k.startswith(prefix)
    }


def format_stats(stats):
    """Given a dictionary following this layout:

        {
            'encoded:label': 'Encoded',
            'encoded:value': 'Yes',
            'encoded:description': 'Indicates if the column is encoded',
            'encoded:include': True,

            'size:label': 'Size',
            'size:value': 128,
            'size:description': 'Size of the table in MB',
            'size:include': True,
        }

    format_stats will convert the dict into this structure:

        {
            'encoded': {
                'id': 'encoded',
                'label': 'Encoded',
                'value': 'Yes',
                'description': 'Indicates if the column is encoded',
                'include': True
            },
            'size': {
                'id': 'size',
                'label': 'Size',
                'value': 128,
                'description': 'Size of the table in MB',
                'include': True
            }
        }
    """
    stats_collector = {}
    for stat_key, stat_value in stats.items():
        stat_id, stat_field = stat_key.split(":")

        stats_collector.setdefault(stat_id, {"id": stat_id})
        stats_collector[stat_id][stat_field] = stat_value

    # strip out all the stats we don't want
    stats_collector = {
        stat_id: stats
        for stat_id, stats in stats_collector.items()
        if stats.get('include', False)
    }

    # we always have a 'has_stats' field, it's never included
    has_stats = {
        'id': 'has_stats',
        'label': 'Has Stats?',
        'value': len(stats_collector) > 0,
        'description': 'Indicates whether there are statistics for this table',
        'include': False,
    }
    stats_collector['has_stats'] = has_stats
    return stats_collector


def unflatten(columns):
    """Given a list of column dictionaries following this layout:

        [{
            'column_comment': None,
            'column_index': Decimal('1'),
            'column_name': 'id',
            'column_type': 'integer',
            'table_comment': None,
            'table_name': 'test_table',
            'table_schema': 'test_schema',
            'table_type': 'BASE TABLE'
        }]

    unflatten will convert them into a dict with this nested structure:

        {
            'test_schema': {
                'test_table': {
                    'metadata': {
                        'comment': None,
                        'name': 'test_table',
                        'type': 'BASE TABLE',
                        'schema': 'test_schema',
                    },
                    'columns': {
                        "id": {
                            'type': 'integer',
                            'comment': None,
                            'index': bigint(1),
                            'name': 'id'
                        }
                    }
                }
            }
        }

    Required keys in each column: table_schema, table_name, column_index

    Keys prefixed with 'column_' end up in per-column data and keys prefixed
    with 'table_' end up in table metadata. Keys without either prefix are
    ignored.
    """
    structured = {}
    for entry in columns:
        schema_name = entry['table_schema']
        table_name = entry['table_name']

        if schema_name not in structured:
            structured[schema_name] = {}
        schema = structured[schema_name]

        if table_name not in schema:
            metadata = get_stripped_prefix(entry, 'table_')
            stats = get_stripped_prefix(entry, 'stats:')
            stats_dict = format_stats(stats)

            schema[table_name] = {
                'metadata': metadata,
                'stats': stats_dict,
                'columns': {}
            }

        table = schema[table_name]

        column = get_stripped_prefix(entry, 'column_')

        # the index should really never be that big so it's ok to end up
        # serializing this to JSON (2^53 is the max safe value there)
        column['index'] = bigint(column['index'])
        table['columns'][column['name']] = column
    return structured


def incorporate_catalog_unique_ids(catalog, manifest):
    nodes = {}

    for schema, tables in catalog.items():
        for table_name, table_def in tables.items():
            unique_ids = manifest.get_unique_ids_for_schema_and_table(
                schema, table_name)

            for unique_id in unique_ids:
                if unique_id in nodes:
                    dbt.exceptions.raise_ambiguous_catalog_match(
                        unique_id, nodes[unique_id], table_def)

                else:
                    table_def_copy = table_def.copy()
                    table_def_copy['unique_id'] = unique_id
                    nodes[unique_id] = table_def_copy

    return nodes


class GenerateTask(CompileTask):
    def _get_manifest(self):
        manifest = dbt.loader.GraphLoader.load_all(self.config)
        return manifest

    def run(self):
        compile_results = None
        if self.args.compile:
            compile_results = super(GenerateTask, self).run()
            if any(r.error is not None for r in compile_results):
                dbt.ui.printer.print_timestamped_line(
                    'compile failed, cannot generate docs'
                )
                return {'compile_results': compile_results}

        shutil.copyfile(
            DOCS_INDEX_FILE_PATH,
            os.path.join(self.config.target_path, 'index.html'))

        adapter = get_adapter(self.config)
        with adapter.connection_named('generate_catalog'):
            manifest = self._get_manifest()

            dbt.ui.printer.print_timestamped_line("Building catalog")
            results = adapter.get_catalog(manifest)

        results = [
            dict(zip(results.column_names, row))
            for row in results
        ]

        nested_results = unflatten(results)
        results = {
            'nodes': incorporate_catalog_unique_ids(nested_results, manifest),
            'generated_at': dbt.utils.timestring(),
        }

        path = os.path.join(self.config.target_path, CATALOG_FILENAME)
        write_json(path, results)

        dbt.ui.printer.print_timestamped_line(
            'Catalog written to {}'.format(os.path.abspath(path))
        )
        # now that we've serialized the data we can add compile_results in to
        # make interpret_results happy.
        results['compile_results'] = compile_results

        return results

    def interpret_results(self, results):
        compile_results = results.get('compile_results')
        if compile_results is None:
            return True

        return super(GenerateTask, self).interpret_results(compile_results)
